from typing import Any
import xarray as xr
import numpy as np    

def retrieve_dBZe(Ze: xr.DataArray, band: str) -> xr.DataArray:
    dBZe = xr.apply_ufunc( lambda x: 10 * np.log10(x), Ze , dask='parallelized')  # convert to dBZ, 10log10(x) = 10 * log10(x)
    dBZe.attrs = {"long_name": f"{band}-band equivalent reflectivity", "units": "dBZe"}
    return dBZe

def retrieve_PhiDP(phiDP: xr.DataArray) -> xr.DataArray:
    phiDP_data = np.rad2deg(
        phiDP
    )  # convert to deg, add -1 because convention is other way around (now the phase shift gets negative, we want it to get positive with range...) TODO: check with Alexander if that makes sense!!
    phiDP_data.attrs = {
        "standard_name": "PhiDP",
        "long_name": "Differential phase shift",
        "units": "deg",
    }
    return phiDP_data

def retrieve_KDP( 
    phiDP: xr.DataArray, moving_windows: tuple[int, int] = (30, 5)
) -> xr.DataArray:
    # time window: timewindow*timeres gives the amount of seconds over which will be averaged
    # calculate KDP from phidp directly
    range_resolution_array = np.diff(phiDP.range)
    if not "time" in phiDP.dims:
        raise ValueError("No time dimension found")
    if len(phiDP.time) < 2:
        raise ValueError(
            "dimension time found to be less than 2. KDP calculation not possible."
        )

    time_window, range_window = moving_windows
    time_rolled_phiDP = phiDP.rolling(
        time=time_window, min_periods=1, center=True
    ).mean()  # moving window average in time
    range_time_rolled_phiDP = time_rolled_phiDP.rolling(
        range=range_window, min_periods=1, center=True
    ).mean()  # moving window average in range
    specific_diff_phase_shift = range_time_rolled_phiDP.diff(dim="range") / (
        2.0 * abs(range_resolution_array) * 1e-3
    )  # in order to get °/km we need to multiply with 1e-3
    specific_diff_phase_shift = specific_diff_phase_shift.reindex(range=phiDP.range, method="nearest")
    specific_diff_phase_shift = specific_diff_phase_shift.rename(
        "specific_diff_phase_shift"
    )
    specific_diff_phase_shift.attrs = {
        "long_name": "Specific differential phase shift",
        "units": "°/km",
    }
    return specific_diff_phase_shift


def retrieve_wind_components(data: xr.Dataset, height: float | list[float] | np.ndarray[Any, np.dtype[np.float64]]) -> xr.Dataset:
    # Calculate the wind components from the Doppler velocities
    ...

def add_all_products_from_LV1(raw: xr.Dataset, band: str) -> xr.Dataset:
    # Add products becomes too big.
    # How far can we go with LV1 data? Is it need to use LV0 from the scratch?
    # Implement here the functions made by Chris
    # It may be insteresting to compare the netcdf from the RPGpy and the one from RPG software
    # There are products that can be only calculated from the spectral data
    data = raw.copy()
    data["dBZe"] = retrieve_dBZe(data["Ze"], band)    
    data["differential_phase"] = retrieve_PhiDP(data["differential_phase"])    
    data["specific_differential_phase"] = retrieve_KDP(data["differential_phase"]) 
    return data

def add_all_products_from_LV0(raw: xr.Dataset) -> xr.Dataset:
    data = raw.copy()       
    chirp_number = xr.DataArray(np.zeros(data.range.size), coords={'range': data.range})   
    for idx, start_chirp_ in enumerate(data['chirp_start_indices'].values):
        chirp_number.loc[{'range': data.range[start_chirp_:]}] = idx
    data['chirp_number'] = chirp_number
    data = data.assign_coords(chirp = np.sort(np.unique(chirp_number.values.astype(int))))    
    # data['doppler_spectrum_v'] = data['doppler_spectrum'] - data['doppler_spectrum_h'] - 2*data['covariance_spectrum_re']
    # VNoisePow = data['integrated_noise'] #Assumption this is the vertical integrated_noise
    # HNoisePow = data['integrated_noise_h']
    # NoiseDensV = VNoisePow/len(data.spectrum)
    # NoiseDensH = HNoisePow/len(data.spectrum)
    # power_noise_h = data['doppler_spectrum_h'].count(dim='spectrum')*NoiseDensH
    # power_noise_v = data['doppler_spectrum_v'].count(dim='spectrum')*NoiseDensV
    # power_signal_v = data['doppler_spectrum_v'].sum(dim='spectrum')
    # power_signal_h = data['doppler_spectrum_h'].sum(dim='spectrum')
    # data['snr_v'] = power_signal_v/power_noise_v
    # data['snr_h'] = power_signal_h/power_noise_h
    # data['sSNR_H'] = data['doppler_spectrum_h'] / NoiseDensH
    # data['sSNR_V'] = data['doppler_spectrum_v'] / NoiseDensV
    # data['ZH'] = data['doppler_spectrum_h'].sum(dim='spectrum')
    # data['ZV'] = data['doppler_spectrum_v'].sum(dim='spectrum')
    # data['ZDR'] = 10*np.log10(data['ZH'])-10*np.log10(data['ZV'])
    # data['ZDP'] = data['ZH']-data['ZV']
    return data


import numpy as np

class BinnedPSD:
    """Binned gamma particle size distribution (PSD).

    Callable class to provide a binned PSD with the given bin edges and PSD
    values.

    Args (constructor):
        The first argument to the constructor should specify n+1 bin edges,
        and the second should specify n bin_psd values.

    Args (call):
        D: the particle diameter.

    Returns (call):
        The PSD value for the given diameter.
        Returns 0 for all diameters outside the bins.
    """

    def __init__(self, bin_edges, bin_psd):
        if len(bin_edges) != len(bin_psd) + 1:
            raise ValueError("There must be n+1 bin edges for n bins.")

        self.bin_edges = bin_edges
        self.bin_psd = bin_psd

    def psd_for_D(self, D):
        if not (self.bin_edges[0] < D <= self.bin_edges[-1]):
            return 0.0

        # binary search for the right bin
        start = 0
        end = len(self.bin_edges)
        while end - start > 1:
            half = (start + end) // 2
            if self.bin_edges[start] < D <= self.bin_edges[half]:
                end = half
            else:
                start = half

        return self.bin_psd[start]

    def __call__(self, D):
        if np.shape(D) == ():  # D is a scalar
            return self.psd_for_D(D)
        else:
            return np.array([self.psd_for_D(d) for d in D])

    def __eq__(self, other):
        if other is None:
            return False
        return (
            len(self.bin_edges) == len(other.bin_edges)
            and (self.bin_edges == other.bin_edges).all()
            and (self.bin_psd == other.bin_psd).all()
        )


#!/usr/bin/env python
# coding: utf-8

from pathlib import Path
from typing import Any
import matplotlib as mpl
from matplotlib import pyplot as plt
import numpy as np
from datetime import datetime
import xarray as xr

from gfatpy.parsivel.plot.utils import DSD_PLOT_INFO
from gfatpy.parsivel.utils import DSD_INFO


def spectrumPlot(
    station: str,
    droplet_spectrum: xr.DataArray,
    date_limits: tuple[datetime, datetime],
    diameter_limits: tuple[float, float] | None = None,
    velocity_limits: tuple[float, float] | None = None,
    output_path: Path | None = None,
    **kwargs,
) -> Path:
    """Generates a plot of the disdrometer spectrum.

    Args:

        - station (str): Station name.
        - droplet_spectrum (list[np.float64]): List of disdrometer spectra.        
        - diameter_limits (tuple[float, float], optional): Diameter axis limits. Defaults to (0, 6.1).
        - velocity_limits (tuple[float, float], optional): Velocity axis limits. Defaults to (0, 12).
        - output_path (Path | None, optional): Directory to save the plot. Defaults to None (current working directory).

    Returns:

        - Path: File path of the generated plot.
    """    

    # Tamaño de la fuente de las etiquetas
    mpl.rc("font", **kwargs)

    # Tamaño de la figura

    # Colormap
    # bounds = DSD_PLOT_INFO['bounds']
    # colors = DSD_PLOT_INFO['colors']
    
    # cm = mpl.colors.ListedColormap(colors)
    # norm = mpl.colors.BoundaryNorm(bounds, cm.N)

    # Mallado del colormap
    # dclasses = np.asarray(DSD_INFO['diameters'],dtype=np.float32)
    # vclasses = np.asarray(DSD_INFO['velocities'],dtype=np.float32)    
    dclasses = droplet_spectrum['dclasses'].values
    vclasses = droplet_spectrum['vclasses'].values

    if diameter_limits is None:
        diameter_limits = (dclasses.min().item(), dclasses.max().item())
    if velocity_limits is None: 
        velocity_limits = (vclasses.min().item(), vclasses.max().item())

    fig, axes = plt.subplots(nrows=1, figsize=(18, 10))
    cmp = droplet_spectrum.plot(x="dclasses", ax=axes, shading="auto") #cmap=cm, norm=norm,
    
    # Etiquetas y rango de visualización
    axes.set_xlabel("Rain-droplet diameter, $m$$m$")
    axes.set_ylabel("Fall velocity, $m/s$")
    breakpoint()
    axes.set_xlim(*diameter_limits)
    axes.set_ylim(*velocity_limits)

    # Dibuja color map    
    cmp.colorbar.set_label("Rain-droplet number concentration, $\#/m^3$")

    # Dibuja el mallado del colormap
    xi, yi = np.meshgrid(dclasses, vclasses)
    axes.plot(xi, yi, "k-", alpha=0.3)  # Dibuja la lineas verticales del mallado
    axes.plot(xi.T, yi.T, "k-", alpha=0.3)  # Dibuja la lineas horizontales del mallado

    d1 = date_limits[0]
    d2 = date_limits[1]
    if d1.date == d2.date:
        period_str = f"{d1.hour} - {d2.hour}"
        date_to_filename = f"{d1.strftime('%y%m%d')}_{d1.strftime('%H%M%S')}-{d2.strftime('%H%M%S')}"
    else: 
        period_str = f"{d1.strftime('%y-%m-%d %H:%M:%S')} - {d2.strftime('%y-%m-%d %H:%M:%S')}"
        date_to_filename = f"{d1.strftime('%y%m%dT%H%M%S')}-{d2.strftime('%y%m%dT%H%M%S')}"

    axes.set_title(f"Disdrometer DSD in {station} | Period: {period_str}")
    
    # Dibuja la linea teórica
    plt.plot(np.asarray(DSD_INFO['gunnKinzer']['diameters']), np.asarray(DSD_INFO['gunnKinzer']['velocities']), "gray")

    if output_path is None:
        output_path = Path.cwd() / f"dsd_spectrum_{date_to_filename}.png"
    
    breakpoint()
    fig.savefig(output_path, bbox_inches="tight", dpi=150)
    plt.close(fig)
    return output_path




# Scattering parameters according to Freudenthaler V., 2015.
from typing import Any
import numpy as np
from scipy import interpolate
import xarray as xr
from pandas.core.arrays import ExtensionArray

from gfatpy.atmo.atmo import attenuated_backscatter, transmittance


Fk = {
    308: 1.05574,
    351: 1.05307,
    354.717: 1.05209,
    355: 1.05288,
    386.890: 1.05166,
    400: 1.05125,
    407.558: 1.05105,
    510.6: 1.04922,
    532: 1.04899,
    532.075: 1.04899,
    607.435: 1.04839,
    710: 1.04790,
    800: 1.04763,
    1064: 1.04721,
    1064.150: 1.04721,
}

epsilon = {
    308: 0.25083,
    351: 0.238815,
    354.717: 0.234405,
    355: 0.23796,
    386.890: 0.23247,
    400: 0.230625,
    407.558: 0.229725,
    510.6: 0.22149,
    532: 0.220455,
    532.075: 0.220455,
    607.435: 0.217755,
    710: 0.21555,
    800: 0.214335,
    1064: 0.212445,
    1064.150: 0.212445,
}

Cs = {
    308: 3.6506e-5,  # K/hPa/m
    351: 2.0934e-5,
    354.717: 2.0024e-5,
    355: 1.9957e-5,
    386.890: 1.3942e-5,
    400: 1.2109e-5,
    407.558: 1.1202e-5,
    510.6: 4.4221e-6,
    532: 3.7382e-6,
    532.075: 3.7361e-6,
    607.435: 2.1772e-6,
    710: 1.1561e-6,
    800: 7.1364e-7,
    1064: 2.2622e-7,
    1064.150: 2.2609e-7,
}

BsT = {
    308: 4.2886e-6,
    351: 2.4610e-6,
    354.717: 2.3542e-6,
    355: 2.3463e-6,
    400: 1.4242e-6,
    510.6: 5.2042e-7,
    532: 4.3997e-7,
    532.075: 4.3971e-7,
    710: 1.3611e-7,
    800: 8.4022e-8,
    1064: 2.6638e-8,
    1064.150: 2.6623e-8,
}

BsC = {
    308: 4.1678e-6,
    351: 2.3949e-6,
    354.717: 2.2912e-6,
    355: 2.2835e-6,
    400: 1.3872e-6,
    510.6: 5.0742e-7,
    532: 4.2903e-7,
    532.075: 4.2878e-7,
    710: 1.3280e-7,
    800: 8.1989e-8,
    1064: 2.5999e-8,
    1064.150: 2.5984e-8,
}

BsC_parallel = {
    308: 4.15052184e-6,
    351: 2.38547616e-06,
    354.717: 2.28368241e-06,
    355: 2.27451222e-06,
    400: 1.38198631e-06,
    510.6: 5.05563542e-07,
    532: 4.27459520e-07,
    532.075: 4.27219387e-07,
    710: 1.32322062e-07,
    800: 8.16989147e-08,
    1064: 2.59074156e-08,
    1064.150: 2.58925276e-08,
}

BsC_perpendicular = {
    308: 1.72550768e-08,
    351: 9.44466863e-09,
    354.717: 8.87554356e-09,
    355: 8.97326485e-09,
    400: 5.28492464e-09,
    510.6: 1.85714694e-09,
    532: 1.56293632e-09,
    532.075: 1.56205831e-09,
    710: 4.73100855e-10,
    800: 2.90465461e-10,
    1064: 9.13006514e-11,
    1064.150: 9.12481844e-11,
}

KbwT = {
    308: 1.01610,
    351: 1.01535,
    354.717: 1.01530,
    355: 1.01530,
    400: 1.01484,
    510.6: 1.01427,
    532: 1.01421,
    532.075: 1.01421,
    710: 1.01390,
    800: 1.01383,
    1064: 1.01371,
    1064.150: 1.01371,
}

KbwC = {
    308: 1.04554,
    351: 1.04338,
    354.717: 1.04324,
    355: 1.04323,
    400: 1.04191,
    510.6: 1.04026,
    532: 1.04007,
    532.075: 1.04007,
    710: 1.03919,
    800: 1.03897,
    1064: 1.03863,
    1064.150: 1.03863,
}

molecular_depolarizationC = {
    351: 0.004158,
    354.717: 0.003959,
    355: 0.003956,
    400: 0.003825,
    510.6: 0.003673,
    532: 0.003656,
    532.075: 0.003656,
    710: 0.003575,
    800: 0.003555,
    1064: 0.003524,
    1064.150: 0.003524,
}

molecular_depolarizationT = {
    351: 0.01559,
    354.717: 0.01554,
    355: 0.01554,
    400: 0.01507,
    510.6: 0.01448,
    532: 0.01441,
    532.075: 0.01441,
    710: 0.01410,
    800: 0.01402,
    1064: 0.01390,
    1064.150: 0.01390,
}

# Create interpolation function once, to avoid re-calculation (does it matter?)
f_ext = interpolate.interp1d(list(Cs.keys()), list(Cs.values()), kind="cubic")
f_bst = interpolate.interp1d(list(BsT.keys()), list(BsT.values()), kind="cubic")
f_bsc = interpolate.interp1d(list(BsC.keys()), list(BsC.values()), kind="cubic")
f_bsc_parallel = interpolate.interp1d(
    list(BsC_parallel.keys()), list(BsC_parallel.values()), kind="cubic"
)
f_bsc_perpendicular = interpolate.interp1d(
    list(BsC_perpendicular.keys()), list(BsC_perpendicular.values()), kind="cubic"
)

# Splines introduce arifacts due to limited input resolution
f_kbwt = interpolate.interp1d(list(KbwT.keys()), list(KbwT.values()), kind="linear")
f_kbwc = interpolate.interp1d(list(KbwC.keys()), list(KbwC.values()), kind="linear")
f_molt = interpolate.interp1d(
    list(molecular_depolarizationT.keys()),
    list(molecular_depolarizationT.values()),
    kind="linear",
)
f_molc = interpolate.interp1d(
    list(molecular_depolarizationC.keys()),
    list(molecular_depolarizationC.values()),
    kind="linear",
)


def molecular_backscatter(
    wavelength: float,
    pressure: np.ndarray[Any, np.dtype[np.float64]],
    temperature: np.ndarray[Any, np.dtype[np.float64]],
    component: str = "total",
) -> np.ndarray[Any, np.dtype[np.float64]]:
    """
    Molecular backscatter calculation.

    Parameters
    ----------
    wavelength : float
       The wavelength of the radiation in air. From 308 to 1064.15
    pressure : float
       The atmospheric pressure. (Pa)
    temperature : float
       The atmospheric temperature. (K)
    component : str
       One of 'total' or 'cabannes'.

    Returns
    -------
    beta_molecular: float
       The molecular backscatter coefficient. (m^-1 * sr^-1)

    References
    ----------
    Freudenthaler, V. Rayleigh scattering coefficients and linear depolarization
    ratios at several EARLINET lidar wavelengths. p.6-7 (2015)
    """
    if component not in [
        "total",
        "cabannes",
        "cabannes_parallel",
        "cabannes_perpendicular",
    ]:
        raise ValueError(
            "Molecular backscatter available only for 'total' or 'cabannes' component."
        )

    if component == "total":
        bs_function = f_bst
    elif component == "cabannes":
        bs_function = f_bsc
    elif component == "cabannes_parallel":
        bs_function = f_bsc_parallel
    elif component == "cabannes_perpendicular":
        bs_function = f_bsc_perpendicular
    else:
        raise ValueError(f"{component} not found.")

    Bs = bs_function(wavelength)

    # Convert pressure to correct units for calculation. (Pa to hPa)
    pressure = pressure / 100.0

    # Calculate the molecular backscatter coefficient.
    beta_molecular = Bs * pressure / temperature

    return beta_molecular


def molecular_lidar_ratio(wavelength: float, component: str = "total") -> float:
    """
    Molecular lidar ratio.

    Parameters
    ----------
    wavelength : float
       The wavelength of the radiation in air. From 308 to 1064.15
    component : str
       One of 'total' or 'cabannes'.

    Returns
    -------
    lidar_ratio_molecular : float
       The molecular backscatter coefficient. (m^-1 * sr^-1)

    References
    ----------
    Freudenthaler, V. Rayleigh scattering coefficients and linear depolarization
    ratios at several EARLINET lidar wavelengths. p.6-7 (2015)
    """
    if component not in ["total", "cabannes"]:
        raise ValueError(
            "Molecular lidar ratio available only for 'total' or 'cabannes' component."
        )

    if component == "total":
        k_function = f_kbwt
    else:
        k_function = f_kbwc

    Kbw = k_function(wavelength)

    lidar_ratio_molecular = 8 * np.pi / 3.0 * Kbw

    return lidar_ratio_molecular


def molecular_depolarization(wavelength: float, component: str = "total") -> float:
    """
    Molecular lidar ratio.

    Parameters
    ----------
    wavelength : float
       The wavelength of the radiation in air. From 308 to 1064.15
    component : str
       One of 'total' or 'cabannes'.

    Returns
    -------
    molecular volume depolarization: float

    References
    ----------
    Freudenthaler, V. Rayleigh scattering coefficients and linear depolarization
    ratios at several EARLINET lidar wavelengths. p.6-7 (2015)
    """

    if component not in ["total", "cabannes"]:
        raise ValueError(
            "Molecular lidar ratio available only for 'total' or 'cabannes' component."
        )

    if component == "total":
        moldepo_function = f_molt
    else:
        moldepo_function = f_molc

    molecular_depolarization = moldepo_function(wavelength)

    return molecular_depolarization


def molecular_extinction(
    wavelength: float,
    pressure: np.ndarray[Any, np.dtype[np.float64]],
    temperature: np.ndarray[Any, np.dtype[np.float64]],
) -> np.ndarray[Any, np.dtype[np.float64]]:
    """
    Molecular extinction calculation.

    Parameters
    ----------
    wavelength : float
       The wavelength of the radiation in air. From 308 to 1064.15
    pressure : float
       The atmospheric pressure. (Pa)
    temperature : float
       The atmospheric temperature. (K)

    Returns
    -------
    alpha_molecular: float
       The molecular extinction coefficient. (m^-1)

    References
    ----------
    Freudenthaler, V. Rayleigh scattering coefficients and linear depolarization
    ratios at several EARLINET lidar wavelengths. p.6-7 (2015)
    """
    cs = f_ext(wavelength)

    # Convert pressure to correct units for calculation. (Pa to hPa)
    pressure = pressure / 100.0

    # Calculate the molecular backscatter coefficient.
    alpha_molecular = cs * pressure / temperature

    return alpha_molecular


def molecular_properties(
    wavelength: float,
    pressure: np.ndarray[Any, np.dtype[np.float64]] | ExtensionArray,
    temperature: np.ndarray[Any, np.dtype[np.float64]] | ExtensionArray,
    heights: np.ndarray[Any, np.dtype[np.float64]] | ExtensionArray,
    times: np.ndarray[Any, np.dtype[np.float64]] | None = None,
    component: str = "total",
) -> xr.Dataset:
    """Molecular Attenuated  Backscatter: beta_mol_att = beta_mol * Transmittance**2

    Args:
        wavelength (float): wavelength of our desired beta molecular attenuated
        pressure (np.ndarray[Any, np.dtype[np.float64]]): pressure profile
        temperature (np.ndarray[Any, np.dtype[np.float64]]): temperature profile
        heights (np.ndarray[Any, np.dtype[np.float64]]): height profile
        times (np.ndarray[Any, np.dtype[np.float64]] | None, optional): time array. Defaults to None. Note: this is not used in the calculation yet. In development.
        component (str, optional): _description_. Defaults to "total".

    Returns:
        xr.Dataset: molecular attenuated backscatter profile, molecular backscatter profile, molecular extinction profile, molecular lidar ratio, molecular depolarization ratio
    """

    # molecular backscatter and extinction #
    beta_mol = molecular_backscatter(
        wavelength, pressure, temperature, component=component
    )
    alfa_mol = molecular_extinction(wavelength, pressure, temperature)
    lr_mol = molecular_lidar_ratio(wavelength, component=component)

    depo_mol = molecular_depolarization(wavelength, component=component)

    """ transmittance """
    transmittance_array = transmittance(alfa_mol, heights)

    """ attenuated molecular backscatter """
    att_beta_mol = attenuated_backscatter(beta_mol, transmittance_array)

    if times is None:
        mol_properties = xr.Dataset(
            {
                "molecular_beta": (["range"], beta_mol),
                "molecular_alpha": (["range"], alfa_mol),
                "attenuated_molecular_beta": (["range"], att_beta_mol),
                "molecular_lidar_ratio": lr_mol,
                "molecular_depolarization": depo_mol,
            },
            coords={"range": heights},
        )
    else:
        mol_properties = xr.Dataset(
            {
                "molecular_beta": (["time", "range"], beta_mol),
                "molecular_alpha": (["time", "range"], alfa_mol),
                "attenuated_molecular_beta": (["time", "range"], att_beta_mol),
                "molecular_lidar_ratio": ([], lr_mol),
                "molecular_depolarization": depo_mol,
            },
            coords={
                "time": times,
                "range": heights,
            },  # FIXME: time is not used in the calculation yet
        )
    return mol_properties




#!/usr/bin/env python
# coding: utf-8

# In[52]:

import os
import glob
import sys
import argparse
import datetime as dt
from optparse import OptionParser
import disdrometer as dd
import parsivelConverter1b as pc1b

__version__ = '1.0.0'
__author__ = 'Juan Antonio Bravo-Aranda'

# script description
BASE_DIR = os.path.dirname(os.path.dirname(__file__))
PROG_NAME = 'disdrometerConverter'
PROG_DESCR = 'converting raw data to netCDF files.'

def main():
    parser = OptionParser(usage="usage %prog [options]",
        version="%prog 1.0")
    parser.add_option("-s", "--station_name",
        action="store",
        dest="station",
        default="UGR",
        help="Measurement station [default: %default].")
    parser.add_option("-d", "--date2plot",
        action="store",
        dest="date2plot",
        default="20191121",
        help="Date to plot [default: %default].")
    parser.add_option("-o", "--dir_out",
        action="store",
        dest="dir_out",
        default=".",
        help="Output folder [default: %default].")
    parser.add_option("-i", "--dir_in",
        action="store",
        dest="dir_in",
        default=".",
        help="Input folder [default: %default].")                
    (options, args) = parser.parse_args()
    dir0 = options.dir_in    
    dir2 = options.dir_out   
    strdate = options.date2plot
    station = options.station
    date_ = dt.datetime.strptime(strdate,'%Y%m%d')
    year = dt.datetime.strftime(date_,'%Y')
    month = dt.datetime.strftime(date_,'%m')
    file0a = os.path.join(dir0, year, month, '%s_%s.mis' % (dt.datetime.strftime(date_,'%Y%m%d'), station))
    if os.path.isfile(file0a):
        print('%s found!' % file0a)    
        file1a = file0a.replace('0a', '1a')
        pc1b.cleanMIS(file0a, file1a)
        
        if os.path.isfile(file1a):
            print('%s found!' % file1a)            
            file1b = file1a.replace('1a', '1b')
            file1b = file1b.replace('mis', 'nc')
            dd.to_nc(file1a, file1b)    
        
            if os.path.isfile(file1b):    
                print('%s successfully converted!' % file1b)
            else:
                print('%s conversion to 1b-level FAILED!' % file0a)
        else:    
            print('%s conversion to 1a-level FAILED!' % file0a)
    else:    
        print('%s not found.' % file0a)        

if __name__== "__main__":
    main()



import numpy as np
import xarray as xr
from numpy.polynomial.polynomial import polyfit, polyval

from gfatpy.utils.optimized import windowed_corrcoefs


def gluing(
    signal_an: xr.DataArray,
    signal_pc: xr.DataArray,
    range_min: float = 1200,
    range_max: float = 5000,
    window_meters: float = 1200,
) -> np.ndarray:
    sel_signal_an = signal_an.sel(
        range=slice(range_min - window_meters / 2, range_max + window_meters / 2)
    )
    sel_signal_pc = signal_pc.sel(
        range=slice(range_min - window_meters / 2, range_max + window_meters / 2)
    )

    ranges = sel_signal_an.range.values
    window_bins = int(window_meters // (ranges[1] - ranges[0]))

    rcs_an = (sel_signal_an * ranges**2).values
    rcs_pc = (sel_signal_pc * ranges**2).values

    corr_coeffs = windowed_corrcoefs(rcs_an, rcs_pc, window_bins)
    gluing_window_bins = np.apply_along_axis(lambda x: x.argmax(), 1, corr_coeffs)
    gluing_bins = gluing_window_bins + int(range_min // (ranges[1] - ranges[0])) + 1

    rcs_an_whole = (signal_an * signal_an.range**2).values
    rcs_pc_whole = (signal_pc * signal_pc.range**2).values
    rcs_gl_whole = rcs_pc_whole.copy()

    for idx in np.arange(rcs_an_whole.shape[0]):
        hw = window_bins // 2  # Half window
        gl_bin = gluing_bins[idx]

        win_an = rcs_an_whole[idx, gl_bin - hw : gl_bin + hw]
        win_pc = rcs_pc_whole[idx, gl_bin - hw : gl_bin + hw]

        reg = polyfit(win_an, win_pc, 1)

        cov = np.corrcoef(win_an, win_pc)[1, 0]
        # print(f"Covariance: {cov}")
        # print(f"Original cov: {corr_coeffs[idx].max()}")
        # print(f"Adjusting analog at profile {idx} to height: {gl_bin * 7.5}")
        if cov > 0.70:
            rcs_gl_whole[idx, :gl_bin] = polyval(rcs_an_whole[idx, :gl_bin], reg)
        else:
            rcs_gl_whole[idx] = np.zeros_like(rcs_gl_whole[idx])

    return rcs_gl_whole / signal_pc.range.values**2



######################################################################################################################################
# this routine automatically reads the binary RPG version 4 files into nc and then resamples them to the common tripex-pol-scan grid
# author: Leonie von Terzi
# last changed: 06.12.2021
######################################################################################################################################

import numpy as np
import xarray as xr
import pandas as pd
import glob
import os
from sys import argv

#import read_RPG_lvl1_v4 as readLV1
import processData as pro
import matplotlib.pyplot as plt

scriptname, date  = argv #, date,dataPath,dataPathOutput,Band #, date, dataPath, dataPathOutput, emptyDataPath
print(date)
#quit()
date = pd.to_datetime(date)
'''
input: 
date: date that you want to have processed
dataPath: path where the X-band data is stored
dataPathOutput: path where to put the plot
emptyDataPath: path to where there is a nc file with empty data in it
'''

pathOut = '/data/optimice/tripex-pol/plots'
filePathW = '/data/obs/campaigns/tripex-pol-scan/wband_scan/l0/'
filePathKa = '/data/obs/site/jue/joyrad35/'
fileExtW = '.LV1.nc'
fileExtKa = '_tripex_pol_II_rhi.znc'
# path to emptyData
emptyDataPath = '/data/optimice/tripex-pol/auxPlotData/noData1.nc'
#inputPath = '/data/obs/campaigns/tripex-pol-scan/wband_scan/l0/'
# This block defines the range reference grid.
# The reference grid is the same for all radars
#
beginRangeRef = 0 # starting height of the ref grid
endRangeRef = 12000 # ending height of the ref grid
rangeFreq = 36 # range resolution of the ref grid
rangeTolerance = 18 # tolerance for detecting the closest neighbour

rangeRef = np.arange(beginRangeRef, endRangeRef, rangeFreq)
# tolerance for detecting closest neighbour (seconds)
timeTolerance = 1
#################################################################################################
# lets first save new binary files into nc files, for all scan patterns, since we want to have all nc files anyway. Since we are first checking if the file is already there, there is no need to change that if we later want to plot just ZEN and had CEL before.

##################################################################################################################
# now lets do the resampling
##################################################################################################################
datenow = pd.Timestamp.now() 

print(date)
if date.strftime('%y%m%d') == datenow.strftime('%y%m%d'):   
  hournow = datenow.strftime('%H')
  hour2proc = np.arange(int(hournow)+1)

else:
  hour2proc = np.arange(24)

# all variables to plot, need to do that separately for now TODO: improve that!
variables = {'Ze':{'vmax':25, 'vmin':-35,'units':'[dB]'},
              'MDV':{'vmax':0, 'vmin':-3,'units':r'[ms$^{-1}$]'},
              'ZDR':{'vmax':0, 'vmin':4,'units':r'[dB]'},
              'KDP':{'vmax':0, 'vmin':4,'units':r'[°km$^{-1}$]'},
              'sZDRmax':{'vmax':0, 'vmin':4,'units':r'[dB]'},
              'WIDTH':{'vmax':0, 'vmin':1,'units':r'[ms$^{-1}$]'},
              'sLDR':{'vmax':-20, 'vmin':-35,'units':r'[dB]'},
              }   
for var in variables:
  fig,axes = plt.subplots(nrows=4,ncols=6,figsize=(24, 12))  
  # fill up the plot according to all the hours that we already had. Therefore select the according axis: 
  for hour in hour2proc:
    if hour in np.arange(6):
      ax = axes[0,hour]
    elif hour in np.arange(6,12):
      ax = axes[1,hour-6]
    elif hour in np.arange(12,18):
      ax = axes[2,hour-12]
    else: 
      ax = axes[3,hour-18]
    timeFreq = '4S'; timeTolerance = '2S'; 
    timeRef = pd.date_range(date,date+pd.offsets.Hour(24)-pd.offsets.Second(1),freq=timeFreq)
  
    year = date.strftime('%Y')
    month = date.strftime('%m')
    day = date.strftime('%d')
  
    dateStr = year+month+day
  
  # defining the input file name
    filesWbandScan = sorted(glob.glob(filePathW+'{year}/{month}/{day}/FIRST_RHI__{date}_{hour}*_P09_{scantype}{fileExt}'.format(year=date.strftime('%Y'),
                                                                              month=date.strftime('%m'),
                                                                              day=date.strftime('%d'),
                                                                              date = date.strftime('%y%m%d'),
                                                                              hour = "{0:0=2d}".format(hour),
                                                                              scantype=scan2proc, # here we just use the ones specified by the bash skript
                                                                              fileExt=fileExtW)))
  
    #print(filesWbandScan)
    
    if filesWbandScan:
      f = filesWbandScan[0]
      wband_scan = xr.open_dataset(f)
              
      wband_scan = pro.mergeChirps(wband_scan)
      wband_scan = wband_scan.rename({'Time':'time'})
      
      wband_scan = pro.calcPhiDP(wband_scan)
      wband_scan = pro.calcKDP(wband_scan,timeWindow=5)
  
      
    
      filesWbandScanLV0 = sorted(glob.glob(filePathW+'{year}/{month}/{day}/FIRST_RHI__{date}_{hour}*_P09_{scantype}{fileExt}'.format(year=date.strftime('%Y'),
                                                                              month=date.strftime('%m'),
                                                                              day=date.strftime('%d'),
                                                                              date = date.strftime('%y%m%d'),
                                                                              hour = "{0:0=2d}".format(hour),
                                                                              scantype=scan2proc, # here we just use the ones specified by the bash skript
                                                                              fileExt='.LV0.nc')))
      
      #wband_scan_LV0 = xr.Dataset()
      f = filesWbandScanLV0[0]
      data = xr.open_dataset(f)
        #
      data = pro.mergeChirps(data,spec=True)
      data = data.rename({'Time':'time'})
      data['sZDR'] = 10*np.log10(data['HSpec']) - 10*np.log10(data['VSpec']) 
      data['sZDRmax'] = data['sZDR'].max(dim='Vel',keep_attrs=True)
          #for t in data.time:
          
      wband_scan = xr.merge([wband_scan, data[['sZDRmax']]])
    # converting Zg to log units
      wband_scan.time.attrs['standard_name'] = 'time'
      wband_scan.time.attrs['long_name'] = 'time'
      wband_scan.time.attrs['units']='seconds since 2001-01-01 00:00:00 UTC'
      wband_scan = xr.decode_cf(wband_scan)
      
      convert = ['Ze']#,'sLDR']#,'ZDR']
      for varconv in convert:
        wband_scan[varconv] = 10*np.log10(wband_scan[varconv])
      
      # now plot the variable in the correct space
      var2plot = (wband_scan[var]).transpose()        
      plot=ax.pcolormesh(wband_scan['range']*np.cos(np.deg2rad(wband_scan['Elv'])), wband_scan['range']*np.sin(np.deg2rad(wband_scan['Elv'])), 
                    var2plot, 
                    vmax = variables[var]['vmax'],
                    vmin = variables[var]['vmin'],
                    cmap='jet')    
      cb = plt.colorbar(plot,ax=ax)
      cb.set_label(var+variables[var]['units'])
      ax.set_ylabel('range [m]')
      ax.set_xlabel('range [m]')
      ax.set_xlim(-10000,10000)
      ax.set_ylim(0,10000)
      #ax.axis('scaled')
      ax.set_title('Hour: {hour}'.format(hour="{0:0=2d}".format(hour)))
      ax.grid()
      plt.tight_layout()
      
  print(var)
  fileName = ('_').join([dateStr,var+'_RHI_W.png'])
  filePathName = ('/').join([pathOut,fileName]) 
  plt.savefig(filePathName,dpi=200,bbox_inches='tight')   
  plt.close()

#############################################################################
# now plot Ka
#############################################################################

print(date)
if date.strftime('%y%m%d') == datenow.strftime('%y%m%d'):   
  hournow = datenow.strftime('%H')
  hour2proc = np.arange(int(hournow)+1)

else:
  hour2proc = np.arange(24)

# all variables to plot, need to do that separately for now TODO: improve that!
variables = {'Zg':{'vmax':25, 'vmin':-35,'units':'[dB]'},
                'VELg':{'vmax':0, 'vmin':-3,'units':r'[ms$^{-1}$]'},
                'RMSg':{'vmax':0, 'vmin':1,'units':r'[ms$^{-1}$]'},
                'LDRg':{'vmax':-20, 'vmin':-35,'units':r'[dB]'},
              }  
for var in variables:
  
  fig,axes = plt.subplots(nrows=4,ncols=6,figsize=(24, 12))  
  # fill up the plot according to all the hours that we already had. Therefore select the according axis: 
  for hour in hour2proc:
    if hour in np.arange(6):
      ax = axes[0,hour]
    elif hour in np.arange(6,12):
      ax = axes[1,hour-6]
    elif hour in np.arange(12,18):
      ax = axes[2,hour-12]
    else: 
      ax = axes[3,hour-18]
    
    timeFreq = '1S'; timeTolerance = '1S'; 
    timeRef = pd.date_range(date,date+pd.offsets.Hour(24)-pd.offsets.Second(1),freq=timeFreq)
  
    year = date.strftime('%Y')
    month = date.strftime('%m')
    day = date.strftime('%d')
  
    dateStr = year+month+day
  
  # defining the input file name
    filesKabandScan = sorted(glob.glob(filePathKa+'{year}/{month}/{day}/{date}_{hour}*{fileExt}'.format(year=date.strftime('%Y'),
                                                                              month=date.strftime('%m'),
                                                                              day=date.strftime('%d'),
                                                                              date = date.strftime('%Y%m%d'),
                                                                              hour = "{0:0=2d}".format(hour),
                                                                              fileExt=fileExtKa)))
  
    
    if filesKabandScan:
      f = filesKabandScan[0]
      KaData = xr.open_dataset(f)
      _, index_time = np.unique(KaData['time'], return_index=True)
      KaData = KaData.isel(time=index_time)
      KaData.time.attrs['units']='seconds since {0}'.format('1970-01-01 00:00:00 UTC')
      KaData = xr.decode_cf(KaData)
      
      convert = ['Zg']#,'sLDR']#,'ZDR']
      for varconv in convert:
        KaData[varconv] = 10*np.log10(KaData[varconv])
      
      # now only plot the first RHI --> select only the first one by looking for monotonically increasing elv
      diff_elv = np.diff(KaData.elv)
      diff_elv = np.insert(diff_elv,-1,0)
      
      time_increasing = KaData.time.where(diff_elv > 0)
      time_increasing = time_increasing.dropna(dim='time')
      data_increasing = KaData.sel(time=time_increasing.values)
      
      # now plot the variable in the correct space
      var2plot = (data_increasing[var]).transpose()  
      elv2plot = np.deg2rad(data_increasing['elv'])
      
      plot=ax.pcolormesh(data_increasing['range']*np.cos(elv2plot),
                          data_increasing['range']*np.sin(elv2plot), 
                          var2plot, 
                          vmax = variables[var]['vmax'],
                          vmin = variables[var]['vmin'],
                          cmap='jet')    
      cb = plt.colorbar(plot,ax=ax)
      cb.set_label(var+variables[var]['units'])
      ax.set_ylabel('range [m]')
      ax.set_xlabel('range [m]')
      ax.set_xlim(-10000,10000)
      ax.set_ylim(0,10000)
      #ax.axis('scaled')
      ax.set_title('Hour: {hour}'.format(hour="{0:0=2d}".format(hour)))
      ax.grid()
      plt.tight_layout()
      
  print(var)
  fileName = ('_').join([dateStr,var+'_RHI_Ka.png'])
  filePathName = ('/').join([pathOut,fileName]) 
  plt.savefig(filePathName,dpi=200,bbox_inches='tight')   
  plt.close()

#####################################################################
# plot difference variables
#####################################################################

print(date)
if date.strftime('%y%m%d') == datenow.strftime('%y%m%d'):   
  hournow = datenow.strftime('%H')
  hour2proc = np.arange(int(hournow)+1)

else:
  hour2proc = np.arange(24)

variables = {'Zg':{'vmax':20, 'vmin':-5, 'name':'DWR','units':'[dB]'},
              'VELg':{'vmax':0.3, 'vmin':-0.3, 'name':'DDV','units':r'[ms$^{-1}$]'},
              'RMSg':{'vmax':0.3, 'vmin':-0.3, 'name':'DSW','units':r'[ms$^{-1}$]'}
            }

  # creating the difference plots
for var in variables.keys():  
    
  fig,axes = plt.subplots(nrows=4,ncols=6,figsize=(24, 12))  
  # fill up the plot according to all the hours that we already had. Therefore select the according axis: 
  for hour in hour2proc:
    if hour in np.arange(6):
      ax = axes[0,hour]
    elif hour in np.arange(6,12):
      ax = axes[1,hour-6]
    elif hour in np.arange(12,18):
      ax = axes[2,hour-12]
    else: 
      ax = axes[3,hour-18]
  
    timeFreq = '1S'; timeTolerance = '1S'; 
    timeRef = pd.date_range(date,date+pd.offsets.Hour(24)-pd.offsets.Second(1),freq=timeFreq)
    
    beginElv = 30 ; endElv = 150 ; elvRef = np.arange(beginElv,endElv+1,3) # we need to go in steps of three here because we have 3 chrip sequences that each take 1 second (so 1° elv, so therefore we do not have a resolution of 1° but rather 3°)
    
    year = date.strftime('%Y')
    month = date.strftime('%m')
    day = date.strftime('%d')
  
    dateStr = year+month+day
  
  # defining the input file name
    filesWbandScan = sorted(glob.glob(filePathW+'{year}/{month}/{day}/FIRST_RHI__{date}_{hour}*_P09_{scantype}{fileExt}'.format(year=date.strftime('%Y'),
                                                                              month=date.strftime('%m'),
                                                                              day=date.strftime('%d'),
                                                                              date = date.strftime('%y%m%d'),
                                                                              hour = "{0:0=2d}".format(hour),
                                                                              scantype=scan2proc, # here we just use the ones specified by the bash skript
                                                                              fileExt=fileExtW)))
  
    #print(filesWbandScan)
    
    if filesWbandScan:
      f = filesWbandScan[2]
      wband_scan_chirp = xr.open_dataset(f)
      wband_scan_chirp.Time.attrs['standard_name'] = 'time'
      wband_scan_chirp.Time.attrs['long_name'] = 'time'
      wband_scan_chirp.Time.attrs['units']='seconds since 2001-01-01 00:00:00 UTC'
      wband_scan_chirp = xr.decode_cf(wband_scan_chirp)
      
      wband_scan = pro.mergeChirps(wband_scan_chirp)
      wband_scan_chirp = wband_scan_chirp.rename({'Time':'time'})
      wband_scan = wband_scan.rename({'Time':'time','Ze':'Zg','MDV':'VELg','WIDTH':'RMSg'})
      
      HourMinW = pd.to_datetime(str(wband_scan.time.values[0])).round(freq='T').strftime('%H%M') # round to closest minute
      
      # now reindex along range:
      wband_scan = wband_scan.reindex({'range':rangeRef},method='nearest',tolerance=rangeTolerance)    
      #wband_scan = wband_scan.reindex({'time':timeRef},method='nearest',tolerance=timeTolerance)
      wband_scan['time'] = wband_scan.Elv
      wband_scan = wband_scan.rename({'time':'elv'})
      wband_scan = wband_scan.drop('Elv')

      _, index_range = np.unique(wband_scan['elv'], return_index=True)
      wband_scan = wband_scan.isel(elv=index_range)
      wband_scan = wband_scan.reindex({'elv':elvRef},method='nearest',tolerance=3)  
      
      #- now the Ka-Band data
      filesKabandScan = sorted(glob.glob(filePathKa+'{year}/{month}/{day}/{date}_{hour}*{fileExt}'.format(year=date.strftime('%Y'),
                                                                              month=date.strftime('%m'),
                                                                              day=date.strftime('%d'),
                                                                              date = date.strftime('%Y%m%d'),
                                                                              hour = "{0:0=2d}".format(hour),
                                                                              fileExt=fileExtKa)))
  
      if filesKabandScan:
        print(HourMinW)
        fKa = []
        for fi in filesKabandScan: # get the same starting minute (Ka file might start at e.g. 14:59 instead of 15:00 minute, so we need to round here. Same as for W-Band)
          DateHourMinKaSec = fi.split(year+'/'+month+'/'+day+'/')[1].split('_')[0]+' '+fi.split(year+'/'+month+'/'+day+'/')[1].split('_')[1]
          HourMinKa = pd.to_datetime(DateHourMinKaSec).round(freq='T').strftime('%H%M')
          if HourMinKa == HourMinW:
            fKa = fi
        
        try:  
          KaData = xr.open_dataset(fKa)
          
          _, index_time = np.unique(KaData['time'], return_index=True)
          KaData = KaData.isel(time=index_time)
          KaData.time.attrs['units']='seconds since {0}'.format('1970-01-01 00:00:00 UTC')
          KaData = xr.decode_cf(KaData)
          KaData = KaData[[var,'elv']]
          KaData = KaData.rename({'elv':'Elv'})
        
          if var == 'Zg':
            KaData[var] = 10*np.log10(KaData[var])
          #KaData[var] = (KaData[var]/KaData[var]).fillna(1)
      
          KaData = KaData.reindex({'range':rangeRef},method='nearest',tolerance=rangeTolerance)    
          #wband_scan = wband_scan.reindex({'time':timeRef},method='nearest',tolerance=timeTolerance)
          KaData['time'] = KaData.Elv
          KaData = KaData.rename({'time':'elv'})
          KaData = KaData.drop('Elv')
          _, index_range = np.unique(KaData['elv'], return_index=True)
          KaData = KaData.isel(elv=index_range)
      
          #- now get the correct Ka-Band angle to go together with the W-Band angle (since we have e.g. 30° C1, 31° C2, 32° C3, all together saved in 30°)
          ChirpMin = [min(wband_scan_chirp.C1Range.values),min(wband_scan_chirp.C2Range.values),min(wband_scan_chirp.C3Range.values)]
          ChirpMax = [max(wband_scan_chirp.C1Range.values),max(wband_scan_chirp.C2Range.values),max(wband_scan_chirp.C3Range.values)]
          ielv = 0
          for ie, el in enumerate(elvRef):
            #print(el)
            #KaSel3 = KaData.sel(elv=el,method='nearest')
            for ielv in range(3):
              KaSel = KaData.sel(elv=(el+ielv),method='nearest')
              #print(KaSel.elv)
              if ielv == 0: # first chirp at beginning of timestep
                KaC1 = KaSel.where((KaSel.range >= ChirpMin[0]),drop=True)
                KaC1 = KaC1.where((KaC1.range <= ChirpMax[0]),drop=True)# and (KaData.range <= ChirpMax[0]))
                KaC1['elv'] = el
          
              elif ielv == 1: #second chirp at middle of time step
                KaC2 = KaSel.where((KaSel.range >= ChirpMin[1]),drop=True)
                KaC2 = KaC2.where((KaC2.range <= ChirpMax[1]),drop=True)# and (KaData.range <= ChirpMax[0]))
                KaC2['elv'] = el
            
              else: # third chirp at end of time step
                KaC3 = KaSel.where((KaSel.range >= ChirpMin[2]),drop=True)
                KaC3 = KaC3.where((KaC3.range <= ChirpMax[2]),drop=True)# and (KaData.range <= ChirpMax[0]))
                KaC3['elv'] = el
            
            
            
            if ie == 0:  
              KaAll = xr.concat([KaC1,KaC2,KaC3],dim='range')
            else:
              KaChirp = xr.concat([KaC1,KaC2,KaC3],dim='range')
              KaAll = xr.concat([KaAll,KaChirp],dim='elv')
            #print(KaAll)
          KaAll = KaAll.reindex({'elv':elvRef},method='nearest',tolerance=3)
                  
        except:
        #if fKa == '/data/obs/site/jue/joyrad35/2021/12/06/20211206_021520_tripex_pol_II_rhi.znc':
          print('no file found')
          KaData = xr.open_dataset(emptyDataPath)
          KaData = KaData.drop('elv')
          KaData = KaData.reindex({'time':wband_scan_chirp.time},method='nearest',tolerance='1s')
          KaData['time'] = wband_scan.elv.values#np.ones(len(KaData.time))*100*np.random_sample(len(KaData.time))
          KaAll = KaData.rename({'time':'elv'})
          KaAll[var] = KaAll[var]*np.nan
            
      #-- now onto the plotting (hopefully):
      diffvar = KaAll[var] - wband_scan[var]
      
      var2plot = (diffvar).transpose()        
      plot=ax.pcolormesh(diffvar['range']*np.cos(np.deg2rad(diffvar['elv'])), diffvar['range']*np.sin(np.deg2rad(diffvar['elv'])), 
                    var2plot, 
                    vmax = variables[var]['vmax'],
                    vmin = variables[var]['vmin'],
                    cmap='jet')    
      cb = plt.colorbar(plot,ax=ax)
      
      cb.set_label(variables[var]['name']+variables[var]['units'])
      ax.set_ylabel('range [m]')
      ax.set_xlabel('range [m]')
      ax.set_xlim(-10000,10000)
      ax.set_ylim(0,10000)
      #ax.axis('scaled')
      ax.set_title('Hour: {hour}'.format(hour="{0:0=2d}".format(hour)))
      ax.grid()
      plt.tight_layout()
      
  print(var)
  fileName = ('_').join([dateStr,variables[var]['name']+'_RHI.png'])
  filePathName = ('/').join([pathOut,fileName]) 
  plt.savefig(filePathName,dpi=200,bbox_inches='tight')   
  plt.close()
       
    
    
    



general_parameters = \
    {'System': '\'MULHACEN\'',
     'Laser_Pointing_Angle': 0,
     'Molecular_Calc': 0,
     'Latitude_degrees_north': 37.164,
     'Longitude_degrees_east': -3.605,
     'Altitude_meter_asl': 680.0,
     'Call sign': "gr"}

channel_parameters = \
{'BT0_p45': {'channel_ID': 1055,
         'Background_Low': 75000,
         'Background_High': 105000,
         'LR_Input': 1,         
         'Depolarization_Factor': 0,
         'Pol_Calib_Range_Min': 1000,
         'Pol_Calib_Range_Max': 3000}, 
 'BC0_p45': {'channel_ID': 1057,
         'Background_Low': 75000,
         'Background_High': 105000,
         'LR_Input': 1,         
         'Depolarization_Factor': 0,
         'Pol_Calib_Range_Min': 1000,
         'Pol_Calib_Range_Max': 3000},
 'BT1_p45': {'channel_ID': 1056,
         'Background_Low': 75000,
         'Background_High': 105000,
         'LR_Input': 1,         
         'Depolarization_Factor': 0,
         'Pol_Calib_Range_Min': 1000,
         'Pol_Calib_Range_Max': 3000},
 'BC1_p45': {'channel_ID': 1058,
         'Background_Low': 75000,
         'Background_High': 105000,
         'LR_Input': 1,         
         'Depolarization_Factor': 0,
         'Pol_Calib_Range_Min': 1000,
         'Pol_Calib_Range_Max': 3000},         
 'BT0_m45': {'channel_ID': 1059,
         'Background_Low': 75000,
         'Background_High': 105000,
         'LR_Input': 1,         
         'Depolarization_Factor': 0,
         'Pol_Calib_Range_Min': 1000,
         'Pol_Calib_Range_Max': 3000},
 'BC0_m45': {'channel_ID': 1061,
         'Background_Low': 75000,
         'Background_High': 105000,
         'LR_Input': 1,         
         'Depolarization_Factor': 0,
         'Pol_Calib_Range_Min': 1000,
         'Pol_Calib_Range_Max': 3000},         
 'BT1_m45': {'channel_ID': 1060,
         'Background_Low': 50000,
         'Background_High': 60000,
         'Laser_Shots': 3000,
         'LR_Input': 1,         
         'Pol_Calib_Range_Min': 1000,
         'Pol_Calib_Range_Max': 3000},
 'BC1_m45': {'channel_ID': 1062,
         'Background_Low': 50000,
         'Background_High': 60000,
         'Laser_Shots': 3000,
         'LR_Input': 1,         
         'Pol_Calib_Range_Min': 1000,
         'Pol_Calib_Range_Max': 3000},         
         }


""" Retrievals of backscatter and extinction based on Raman measurements

.. warning::
   These functions have not been tested!
"""
from pdb import set_trace
from typing import Tuple
from loguru import logger
import numpy as np
from scipy.integrate import cumulative_trapezoid
from scipy.signal import savgol_filter

from gfatpy.atmo.atmo import number_concentration_from_meteo, transmittance
from gfatpy.atmo.rayleigh import molecular_properties
from gfatpy.lidar.utils.utils import signal_to_rcs


def retrieve_extinction_deprecated(
    signal: np.ndarray,
    ranges: np.ndarray,
    wavelengths: Tuple[float, float],
    pressure: np.ndarray,
    temperature: np.ndarray,
    window_size_m: float = 100.0,
    savgol_order: int = 1,
    particle_angstrom_exponent: float = 1,
) -> np.ndarray:
    """Retrieve particle extinction from lidar signal.

    This function calculates the particle extinction from the lidar signal using
    range corrected signal (rcs) and molecular properties.

    Args:
       signal (np.ndarray): Lidar signal.
       ranges (np.ndarray): Array of range values corresponding to the signal.
       wavelengths (Tuple[float, float]): Tuple of two wavelengths (in nm).
       pressure (np.ndarray): Atmospheric pressure profile.
       temperature (np.ndarray): Atmospheric temperature profile.
       window_size_m (float, optional): Window size for derivative calculation in meters.
       savgol_order (int, optional): Order of Savitzky-Golay filter.
       particle_angstrom_exponent (float, optional): Particle Ångström exponent.

    Returns:
       Union[np.ndarray, None]: Array containing the particle extinction values.
       If the number density or signal is invalid, the corresponding element in
       the result will be set to np.nan. If the input signal and number density
       are both invalid, the function returns None.
    """

    # Retrieve range corrected signal
    rcs = signal_to_rcs(signal, ranges)

    elastic_rayleigh_beta = molecular_properties(wavelengths[0], pressure, temperature, heights=ranges)
    raman_rayleigh_beta = molecular_properties(wavelengths[1], pressure, temperature, heights=ranges)
    elastic_alpha_molecular = elastic_rayleigh_beta['molecular_alpha']
    raman_alpha_molecular = raman_rayleigh_beta['molecular_alpha']

    # Retrieve number density
    number_density = number_concentration_from_meteo(pressure, temperature, atmospheric_component='nitrogen')

    # Calculate ratio
    ratio = np.nan * np.ones(signal.shape)
    valid_idx = np.logical_and(number_density > 0, signal > 0)
    ratio[valid_idx] = np.ma.log(number_density[valid_idx] / rcs[valid_idx])

    # Apply derivative
    dz = np.median(np.diff(ranges))  # type: ignore
    window_size_bin = np.floor(window_size_m / dz).astype(int)
    
    derivative = savgol_filter(
        ratio,
        window_size_bin,
        savgol_order,
        deriv=1,
        delta=dz,
        mode="nearest",
        cval=np.nan,
    )  # Calculate 1st derivative

    # Calculate particle extinction
    cte = 1 + (wavelengths[1] / wavelengths[0]) ** (-particle_angstrom_exponent)
    particle_alpha = (
        derivative - raman_alpha_molecular - elastic_alpha_molecular
    ) / cte

    # Fill like overlap
    idx_overlap = np.ceil(window_size_bin / 2).astype(int)
    particle_alpha[:idx_overlap] = particle_alpha[idx_overlap]

    return particle_alpha


def retrieve_backscatter(
    signal_raman: np.ndarray,
    signal_emission: np.ndarray,
    extinction_profile: np.ndarray,
    range_profile: np.ndarray,
    wavelengths: Tuple[float, float],
    pressure: np.ndarray,
    temperature: np.ndarray,
    particle_angstrom_exponent: float = 1,
    reference: Tuple[float, float] = (8000, 8500),
    beta_part_ref: float = 0,
) -> np.ndarray:
    """Calculates the aerosol backscatter coefficient.

    This function calculates the aerosol backscatter coefficient based on preprocessed
    elastic and Raman signals and the retrieved aerosol extinction coefficient.

    Parameters:
       signal_raman (np.ndarray): The range-corrected Raman signal (1D array of size M).
       signal_emission (np.ndarray): The range-corrected elastic signal (at the emission wavelength, 1D array of size M).
       extinction_profile (np.ndarray): The aerosol extinction coefficient at each point of the signal profile (1D array of size M).
       range_profile (np.ndarray): Array of range values corresponding to the signal (1D array of size M).
       wavelengths (Tuple[float, float]): Tuple of two wavelengths (in nm) for the emission and Raman signals.
       pressure (np.ndarray): Atmospheric pressure profile (1D array of size M, [Pa]).
       temperature (np.ndarray): Atmospheric temperature profile (1D array of size M, [K]).
       particle_angstrom_exponent (float, optional): Particle Ångström exponent (default: 1).
       reference (Tuple[float, float], optional): Reference altitude range (default: (8000, 8500)).
       beta_aer_ref (float, optional): The molecular backscatter coefficient at the reference altitude (default: 0).

    Returns:
       np.ndarray: The aerosol backscatter coefficient [m^-1].

    Notes:
       The aerosol backscatter coefficient is given by the formula:
       β_aer(R, λ0) = [β_aer(R0, λ0) + β_mol(R0, λ0)] * (P(R0, λ_ra) * P(R, λ0)) / (P(R0, λ0) * P(R, λ_ra)) *
                      exp(-∫(R0, R) [α_aer(r, λ_ra) + α_mol(r, λ_ra)] dr) / exp(-∫(R0, R) [α_aer(r, λ0) + α_mol(r, λ0)] dr) -
                      β_mol(R, λ0)

    References:
       Ansmann, A. et al. Independent measurement of extinction and backscatter profiles in cirrus clouds by using a combined Raman elastic-backscatter lidar.
       Applied Optics Vol. 31, Issue 33, pp. 7113-7131 (1992)
    """
    idx_ref = np.logical_and(
        range_profile >= reference[0], range_profile <= reference[1]
    )
    if not idx_ref.any():
        raise ValueError("Range `reference` out of rcs size.")

    # Calculate profiles of molecular extinction
    atmo_data = molecular_properties(
        wavelengths[0], pressure, temperature, heights=range_profile
    )
    beta_mol = atmo_data["molecular_beta"].values
    alpha_mol_emmision = atmo_data["molecular_alpha"].values
    atmo_data_raman = molecular_properties(
        wavelengths[1], pressure, temperature, heights=range_profile
    )
    beta_mol_raman = atmo_data_raman["molecular_beta"].values
    alpha_mol_raman = atmo_data_raman["molecular_alpha"].values
    
    #Remove nan and negative values from extiction
    extinction_profile = np.nan_to_num(extinction_profile)
    extinction_profile[extinction_profile < 0] = 0
    logger.info('Negative values in extinction profile were removed.')
    
    alpha_part_raman = extinction_profile * (wavelengths[1] / wavelengths[0]) ** (
        -particle_angstrom_exponent
    )

    T_raman = np.exp(-cumulative_trapezoid(alpha_mol_raman + alpha_part_raman, range_profile, initial=0))  # type: ignore
    T_elastic = np.exp(-cumulative_trapezoid(alpha_mol_emmision + extinction_profile, range_profile, initial=0))  # type: ignore

    # Calibration at reference altitude
    beta_at_reference = beta_part_ref + np.nanmean(beta_mol[idx_ref])
    gain_ratio = ( (signal_emission / signal_raman) * (T_raman / T_elastic) * (beta_mol_raman / beta_at_reference) )
    gain_ratio_at_reference = np.nanmean(gain_ratio[idx_ref])
    beta_part = (
        beta_mol_raman
        * (1 / gain_ratio_at_reference)
        * ((signal_emission * T_raman) / (signal_raman * T_elastic))
        - beta_mol
    )
    return beta_part

def retrieve_extinction(
    signal: np.ndarray,
    ranges: np.ndarray,
    wavelengths: Tuple[float, float],
    pressure: np.ndarray,
    temperature: np.ndarray,
    particle_angstrom_exponent: float = 1,
    reference: Tuple[float, float] = (8000, 8500),
    **kwargs: dict,
) -> np.ndarray:
    """Retrieve particle extinction from lidar signal.

    Args:
        signal (np.ndarray): Raman lidar signal.
        ranges (np.ndarray): Array of range values corresponding to the signal.
        wavelengths (Tuple[float, float]): Tuple of elastic and Raman wavelengths (in nm).
        pressure (np.ndarray): Atmospheric pressure profile.
        temperature (np.ndarray): Atmospheric temperature profile.
        particle_angstrom_exponent (float, optional): Particle extinction-related Ångström exponent. Defaults to 1.
        reference (Tuple[float, float], optional): Aerosol-partile free reference region. Defaults to (8000, 8500).

    Raises:
        ValueError: If `reference` is out of range. 

    Returns:
        np.ndarray: Particle extinction profile.
    """    
    # Retrieve range corrected signal
    rcs = signal_to_rcs(signal, ranges)

    elastic_rayleigh_beta = molecular_properties(wavelengths[0], pressure, temperature, heights=ranges)
    raman_rayleigh_beta = molecular_properties(wavelengths[1], pressure, temperature, heights=ranges)
    elastic_alpha_molecular = elastic_rayleigh_beta['molecular_alpha']
    raman_alpha_molecular = raman_rayleigh_beta['molecular_alpha']
    transmittance_twice = transmittance(elastic_alpha_molecular + raman_alpha_molecular, ranges)
    
    #Normalize rcs to attenuated molecular backscatter (elastic and raman)
    attenuated_raman_molecular_backscatter = raman_rayleigh_beta['molecular_beta']*transmittance_twice
    idx_ref = np.logical_and(ranges >= reference[0], ranges <= reference[1])
    if not idx_ref.any():
        raise ValueError("Range `reference` out of rcs size.")
    reference_value = np.nanmean(rcs[idx_ref])
    reference_att_mol_beta = np.nanmean(attenuated_raman_molecular_backscatter[idx_ref])
    rcs2attbeta =reference_att_mol_beta*( rcs / reference_value )
    att_beta_ratio = rcs2attbeta / attenuated_raman_molecular_backscatter
    spectral_dependence = 1 + (wavelengths[1] / wavelengths[0]) ** (-particle_angstrom_exponent)
    aod = -np.log(att_beta_ratio)/spectral_dependence

    #extinction as slope of aod
    extincion = np.gradient(aod, ranges)

    if 'full_overlap_height' in kwargs:
        full_overlap_height = kwargs['full_overlap_height']
        idx_full = np.argmin(np.abs(ranges - full_overlap_height))
        extincion[:idx_full] = extincion[idx_full]

    #Plot in two axes: aod and extinction
    # import matplotlib.pyplot as plt
    # fig, ax = plt.subplots(1, 2, figsize=(10, 5))
    # ax[0].plot(aod, ranges)
    # ax[0].set_xlabel('AOD')
    # ax[1].plot(1e6*extincion, ranges)
    # fig.savefig('aod.png')
    return extincion



import numpy as np
from numpy.polynomial.polynomial import polyfit

import numba

from gfatpy.utils.utils import linear_fit


def best_slope_fit(mat1: np.ndarray, mat2: np.ndarray, window: int) -> np.ndarray:

    assert mat1.shape == mat2.shape, "Matrices shape must match"
    assert isinstance(window, int), "Window argument must be an integer"

    x = np.arange(window)
    res = np.array([])

    # Iter in first dimension, equivalent to time in lidar case
    for idx in range(mat1.shape[0]):
        windowed1 = rolling(mat1[idx], window)
        windowed2 = rolling(mat2[idx], window)

        slopes1 = polyfit(x, windowed1.T, 1)[0]
        slopes2 = polyfit(x, windowed2.T, 1)[0]
        chosen_group = np.argmin(np.abs(slopes1 - slopes2) / slopes1)
        res = np.hstack([res, chosen_group + np.floor(window / 2)])  # Append chosen bin

    return res.astype(int)


@numba.njit(parallel=True)
def windowed_corrcoefs(arr1: np.ndarray, arr2: np.ndarray, w_size: int):
    range_shape = arr1.shape[1] - (w_size - 1)
    _corrcoefs = np.empty((arr1.shape[0], range_shape))
    for t_idx in numba.prange(arr1.shape[0]):
        w1 = rolling(arr1[t_idx], w_size)
        w2 = rolling(arr2[t_idx], w_size)
        for idx in numba.prange(range_shape):
            _w1 = w1[idx]
            _w2 = w2[idx]
            coeff = np.corrcoef(_w1, _w2)[1, 0]
            _corrcoefs[t_idx][idx] = coeff
    return _corrcoefs


@numba.njit(parallel=True)
def windowed_proportional(arr1: np.ndarray, arr2: np.ndarray, /, *, w_size: int):

    assert arr1.shape == arr2.shape, "Matrices shape must match"
    # assert isinstance(w_size, int), "Window argument must be an integer"

    range_shape = arr1.shape[1] - (w_size - 1)
    _proportional = np.full((arr1.shape[0], range_shape), np.nan)
    _factor = np.full((arr1.shape[0], range_shape), np.nan)

    for t_idx in numba.prange(arr1.shape[0]):
        w1 = rolling(arr1[t_idx], w_size)
        w2 = rolling(arr2[t_idx], w_size)

        for idx in numba.prange(range_shape):
            _w1 = w1[idx]
            _w2 = w2[idx]

            ratio = np.mean(_w2 / _w1)
            adj = _w1 * ratio
            _factor[t_idx][idx] = ratio
            _proportional[t_idx][idx] = (np.abs(adj - _w2) / _w2).mean()

    return _factor, _proportional


@numba.njit()
def rolling(a, window):
    shape = (a.size - window + 1, window)
    strides = (a.itemsize, a.itemsize)
    return np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)


@numba.njit
def correlate_vector_to_matrix(vector, matrix):
    """Correlate a vector to a matrix

    Args:
        vector (np.ndarray): Vector to correlate
        matrix (np.ndarray): Matrix to correlate

    Returns:
        np.ndarray: Correlation result
    """
    corr = np.zeros(matrix.shape[1])
    for i in range(matrix.shape[1]):
        corr[i] = np.corrcoef(vector, matrix[:, i])[0, 1]
    return corr


def rolling_window_test(a, window):
    shp = a.shape[:-1] + (a.shape[-1] - window + 1, window)
    strides = a.strides + (a.strides[-1],)
    return np.lib.stride_tricks.as_strided(a, shape=shp, strides=strides)

def moving_linear_fit(x_array, y_array, window_size, **kwargs):
    """ Moving polyfit 
    """
    xdata = rolling_window_test(x_array, window_size).T
    ydata = rolling_window_test(y_array, window_size).T
    
    d, slope = np.nan*np.ones(len(x_array)), np.nan*np.ones(len(x_array))
    mrse, anderson_coef = np.nan*np.ones(len(x_array)), np.nan*np.ones(len(x_array))
    for idx in range(ydata.shape[1]):
        # fit_parameters = np.polyfit(xdata[:,idx], ydata[:,idx], deg = degree, full=False)
        stats = linear_fit(xdata[:,idx], ydata[:,idx])
        slope[idx]  = stats["parameters"][1] 
        std_slope = stats["standard_deviation_parameters"][1]
        d[idx] = stats["durbin_watson"]
        mrse[idx] = stats["msre"]
        anderson_coef[idx] = stats["anderson"][0]
        # if idx in  [1100, 1300, 1600, 1900]:
        #     import matplotlib.pyplot as plt
        #     fig, ax = plt.subplots()
        #     ax.plot(xdata[:,idx], ydata[:,idx], linewidth=0, marker='o', label=f'{ranges[idx]}')
        #     ax.plot(xdata[:,idx],np.polyval(np.flip(parameters), xdata[:,idx]), label='fit')
        #     fig.savefig(f'test_dws_{idx}.png', dpi=300)
        #     plt.close(fig)
    results = {"slope": slope, "std_slope": std_slope ,"durbin_watson": d, "mrse": mrse, "anderson": anderson_coef}
    return results



general_parameters = \
    {'System': '\'MULHACEN\'',
     'Laser_Pointing_Angle': 0,
     'Molecular_Calc': 0,
#     'Latitude_degrees_north': 37.164,
#     'Longitude_degrees_east': -3.605,
#     'Altitude_meter_asl': 680.0,
     'Call sign': "gr"}

#BT0: 532xpa - 1284
#BC0: 532xpp - 
#BT1: 532xca - 1283
#BC1: 532xcp
#BT2: 355xta - 1282
#BC2: 355xtp
#BT3: 1064xta - 1045
#BC3: 530xtp - 1048
#BC4: 353xtp - 1047
#BC5: 408xtp

channel_parameters = \
    {'BT0': {'channel_ID': 1284,
                    'Background_Low': 75000,
                    'Background_High': 105000,
                    'LR_Input': 1, },
     'BT1': {'channel_ID': 1283,
                    'Background_Low': 75000,
                    'Background_High': 105000,
                    'LR_Input': 1, },
     'BT2': {'channel_ID': 1282,
                    'Background_Low': 75000,
                    'Background_High': 105000,
                    'LR_Input': 1, },
     'BT3': {'channel_ID': 1045,
                    'Background_Low': 75000,
                    'Background_High': 105000,
                    'LR_Input': 1, },               
     'BC0': {'channel_ID': 1094,
                    'Background_Low': 75000,
                    'Background_High': 105000,
                    'LR_Input': 1, },               
     'BC1': {'channel_ID': 1093,
                    'Background_Low': 75000,
                    'Background_High': 105000,
                    'LR_Input': 1, },               
     'BC2': {'channel_ID': 1090,
                    'Background_Low': 75000,
                    'Background_High': 105000,
                    'LR_Input': 1, },               
     'BC4': {'channel_ID': 1047,
                    'Background_Low': 75000,
                    'Background_High': 105000,
                    'LR_Input': 1, },               

     }



general_parameters = \
    {'System': '\'MULHACEN\'',
     'Laser_Pointing_Angle': 0,
     'Molecular_Calc': 0,
#     'Latitude_degrees_north': 37.164,
#     'Longitude_degrees_east': -3.605,
#     'Altitude_meter_asl': 680.0,
     'Call sign': "gr"}

channel_parameters = \
    {'BT0': {'channel_ID': 1284,
#                    'channel_str_ID': 'gr_0532xpa_elPR',
                    'Background_Low': 75000,
                    'Background_High': 105000,
                    'LR_Input': 1, },
     'BT1': {'channel_ID': 1283,
#                    'channel_str_ID': 'gr_0532xca_elPT',
                    'Background_Low': 75000,
                    'Background_High': 105000,
                    'LR_Input': 1, },
     'BT2': {'channel_ID': 1282,
#                    'channel_str_ID': 'gr_0355xta_elT',
                    'Background_Low': 75000,
                    'Background_High': 105000,
                    'LR_Input': 1, },
     'BT3': {'channel_ID': 1045,
#                    'channel_str_ID': 'gr_01064xta',
                    'Background_Low': 75000,
                    'Background_High': 105000,
                    'LR_Input': 1, },
     }



import os
import time
import numpy as np
import pandas as pd
import xarray as xr
import glob as glob
from sys import argv
#from netCDF4 import Dataset
import math
import matplotlib.pyplot as plt
from datetime import date
import time
import multiprocessing
import warnings
warnings.filterwarnings("ignore") # TODO don't use normally!!!
def mergeChirps(data,spec=False):
# this function merges the chirps of the dataset. I am doing this, because the non-polarimetric X and Ka-Band don't have a chirps and therefore we only have one fixed range-resolution
# input: data as xarray-dataset, if you set spec=True, the spectral data will be merged.
    try:
    	chirpNum = data.ChirpNum.values[0]
    except:
    	chirpNum = data.ChirpNum.values
    if spec==True:
        #try:
        #print(len(data.DoppLen))
        if len(data.DoppLen)>3:
        	maxVel = data.MaxVel.values[0]; doppLen = data.DoppLen.values[0] 
        else:	
        	maxVel = data.MaxVel.values; doppLen = data.DoppLen.values 
        dv_vec = np.empty(chirpNum)
        maxRange = [max(data.C1Range),max(data.C2Range),max(data.C3Range)]
        for chirp in range(chirpNum):
            ChRange = 'C{chirp}Range'.format(chirp=chirp+1)
            ChHSpec = 'C{chirp}HSpec'.format(chirp=chirp+1)
            ChVSpec = 'C{chirp}VSpec'.format(chirp=chirp+1)
            ChReVHSpec = 'C{chirp}ReVHSpec'.format(chirp=chirp+1)
            ChHNoisePow = 'C{chirp}HNoisePow'.format(chirp=chirp+1)
            ChVNoisePow = 'C{chirp}VNoisePow'.format(chirp=chirp+1)
            ChVel = 'C{chirp}Vel'.format(chirp=chirp+1)
            
            #- calculate noise density:
            NoiseDensV = data[ChVNoisePow]/doppLen[chirp]
            NoiseDensH = data[ChHNoisePow]/doppLen[chirp]
            #- now we need to decompose the VSpec, because this is not actually the vert. Spectrum but saved as a composite of H and V
            data[ChVSpec] = 4*data[ChVSpec] - data[ChHSpec] - 2*data[ChReVHSpec]
            #- there was a software mistake, so the noise of ReHV was not stored correctly. Therefore Alexander suggested to use s SNR threshold of 10dB, otherwise the data will be masked. For this we need to calculate SNR: SNR = signal power/noise power. In order to calculate the correct values, we need to mask all values below -90 dBZ.
            specThreshold = 10**(-90/10)
            data[ChHSpec] = data[ChHSpec].where(data[ChHSpec]>specThreshold,np.NaN)
            data[ChVSpec] = data[ChVSpec].where(data[ChVSpec]>specThreshold,np.NaN)
            NoisePowH = data[ChHSpec].count(dim=ChVel)*NoiseDensH
            NoisePowV = data[ChVSpec].count(dim=ChVel)*NoiseDensV
            SignalPowH = data[ChHSpec].sum(dim=ChVel)
            SignalPowV = data[ChVSpec].sum(dim=ChVel)
            data['SNR_H'] = SignalPowH / NoisePowH
            data['SNR_V'] = SignalPowV / NoisePowV
            #-- now also make it spectral (so spectral ZDR against SNR in respective bin (we see also some high ZDR values at the left (fast) edge, which should not be there)) and then the data should be masked with sSNR > 10dB aswell
            data['sSNR_H'] = data[ChHSpec] / NoiseDensH
            data['sSNR_V'] = data[ChVSpec] / NoiseDensH 
            #-- it is the easiest to calculate Z,ZDR,ZDP at this stage, because once the chirps are merged, it is difficult to sort out the dv needed to integrate over the spectrum...
            data['ZH'] = data[ChHSpec].sum(dim=ChVel)
            data['ZV'] = data[ChVSpec].sum(dim=ChVel)
            data['ZDR'] = 10*np.log10(data['ZH'])-10*np.log10(data['ZV'])
            data['ZDP'] = data['ZH']-data['ZV']
            #- because the different chirps have a different Doppler resolution, we also need to regrid along that axis:
            velData = np.linspace(-maxVel[chirp], maxVel[chirp], doppLen[chirp], dtype=np.float32) # in the dataformat, the dopplervelocity was not assigned yet, but rather it is stored as maxVel and doppLen which you then need to manually assign to the doppler Vel coordinate
            dv_diff = np.diff(velData) # since we regrid along the doppler axis, we need to divide the regridded doppler spectra by dv
            #print(dv_diff)
            dv_vec[chirp] = dv_diff[0]
            data = data.assign({ChVel:velData})
            velRef = np.linspace(-maxVel[0], maxVel[0], doppLen[0], dtype=np.float32)# just use the Doppler velocity from the smallest chirp
            data = data.reindex({ChVel:velRef}, method = 'nearest', tolerance = 0.05) # regrid
            data[ChVSpec] = data[ChVSpec]/dv_vec[chirp]
            data[ChHSpec] = data[ChHSpec]/dv_vec[chirp]
            data[ChReVHSpec] = data[ChReVHSpec]/dv_vec[chirp]
            #- now we can rename the variables to without the chirps and then we merge the datasets along the range coordinate
            dataCh = data[[ChVSpec,ChHSpec,ChHNoisePow,ChVNoisePow,'ZH','ZV','ZDR','ZDP','SNR_H','SNR_V','sSNR_H','sSNR_V']]
            dataCh = dataCh.rename({ChRange:'range',ChVSpec:'VSpec',ChHSpec:'HSpec',ChHNoisePow:'HNoisePow',ChVNoisePow:'VNoisePow',ChVel:'Vel'})
            if chirp==0:
                finalData = dataCh
            else:
                finalData = xr.concat([finalData,dataCh],dim='range')
        dv =  xr.DataArray(dv_vec,dims=('Chirp'))
        temp = xr.Dataset({'Azm':data.Azm,
                       'Elv':data.Elv,
                       'RangeRes':data.RangeRes,
                       #'delRange':delRange,
                       'dv':dv,
                       'maxVel':maxVel,
                       'doppLen':doppLen,
                       'maxRange':maxRange
                       })
    else:
        for chirp in range(chirpNum):
            ChRange = 'C{chirp}Range'.format(chirp=chirp+1)
            ChZDR = 'C{chirp}ZDR'.format(chirp=chirp+1)
            ChPhiDP = 'C{chirp}PhiDP'.format(chirp=chirp+1)
            ChRHV = 'C{chirp}RHV'.format(chirp=chirp+1)
            ChZe = 'C{chirp}ZE'.format(chirp=chirp+1)
            ChSkew = 'C{chirp}Skew'.format(chirp=chirp+1)
            ChWidth = 'C{chirp}SpecWidth'.format(chirp=chirp+1)
            ChKurt = 'C{chirp}Kurt'.format(chirp=chirp+1)
            ChVel = 'C{chirp}MeanVel'.format(chirp=chirp+1)
            # now merge the dataset along the range coordinate
            dataCh = data[[ChZDR,ChPhiDP,ChRHV,ChZe,ChSkew,ChWidth,ChKurt,ChVel]]
            dataCh = dataCh.rename({ChRange:'range',ChZDR:'ZDR',ChPhiDP:'PhiDP',ChRHV:'RHV',ChZe:'DBZ',ChSkew:'SK',ChWidth:'WIDTH',ChKurt:'KURT',ChVel:'VEL'})
            if chirp==0:
                finalData = dataCh    
            else:
                finalData = xr.concat([finalData,dataCh],dim='range')
    
        delRange = np.concatenate((selfDiv(data.C1Range)*data.RangeRes[0],
                           selfDiv(data.C2Range)*data.RangeRes[1],
                           selfDiv(data.C3Range)*data.RangeRes[2],))
        delRange = xr.DataArray(delRange,
                            dims=('range'),
                            coords={'range':finalData.range})
        temp = xr.Dataset({'Azm':data.Azm,
                       'Elv':data.Elv,
                       'RangeRes':data.RangeRes,
                       'delRange':delRange})
    
    finalData = xr.merge([finalData, temp])
    return finalData 


def dealiazeParallel(data):
	
	n_cores = 16#multiprocessing.cpu_count()
	#if n_cores > 1:
	#	n_cores = n_cores - 1 # we have the main function running on one core and our institute does not allow to use all cores
	pool = multiprocessing.Pool(n_cores)
	Nyquvist = data.maxVel.values#[6.350626, 4.981658, 2.61496353]
	RangeEnd = data.maxRange.values#[715.4785, 5902.698, 17994.29]
	doppLen = data.doppLen.values
	vel = data.Vel.values
	dv = np.diff(vel)[0]
	newVel = np.arange(-Nyquvist[0]*2,Nyquvist[0]*2+dv,dv,dtype=np.float32)
			
	data['minVelH'] = xr.DataArray(dims=('range','Time'),
			                				coords={'range':data.range,'Time':data.Time})
	data['maxVelH'] = data['minVelH'].copy()
	data['minVelV'] = data['minVelH'].copy()
	data['maxVelV'] = data['minVelH'].copy()
	data['minVelZDR'] = data['minVelH'].copy()
	data['maxVelZDR'] = data['minVelH'].copy()
	data['HSpec_rot'] = xr.DataArray(dims=('Time','range','vel_rot'),coords={'vel_rot':newVel,'Time':data.Time,'range':data.range})
	data['VSpec_rot'] = data['HSpec_rot'].copy()
	data['sZDR_rot'] = data['HSpec_rot'].copy()
	#- testing dealiasing:
	for tind,t in enumerate(data.Time):
		print(t.values)
		#print(data)
		dataRangeValues = data.range.where(np.isfinite(10*np.log10(data.sel(Time=t).ZH)),drop=True)
		args =  [(r,data.sel(Time=t)) for r in dataRangeValues]
		for r,HSpec,VSpec,sZDR,sSNR_H,sSNR_V,maxVelH,minVelH,maxVelV,minVelV,maxVelZDR,minVelZDR,HSpec_rot,VSpec_rot,sZDR_rot in pool.starmap(dealiazeOneHeight,args):
			data['HSpec'].loc[t,r,:] = HSpec
			data['VSpec'].loc[t,r,:] = VSpec
			data['sZDR'].loc[t,r,:] = sZDR
			data['sSNR_H'].loc[t,r,:] = sSNR_H
			data['sSNR_V'].loc[t,r,:] = sSNR_V
			data['maxVelH'].loc[r,t] = maxVelH
			data['minVelH'].loc[r,t] = minVelH
			data['maxVelV'].loc[r,t] = maxVelV
			data['minVelV'].loc[r,t] = minVelV
			data['maxVelZDR'].loc[r,t] = maxVelZDR
			data['minVelZDR'].loc[r,t] = minVelZDR
			data['HSpec_rot'].loc[t,r,:] = HSpec_rot
			data['VSpec_rot'].loc[t,r,:] = VSpec_rot
			data['sZDR_rot'].loc[t,r,:] = sZDR_rot
		
	return data

def dealiazeOneHeight(r,data):
	Nyquvist = data.maxVel.values#[6.350626, 4.981658, 2.61496353]
	RangeEnd = data.maxRange.values#[715.4785, 5902.698, 17994.29]
	doppLen = data.doppLen.values
	vel = data.Vel.values
	dv = np.diff(vel)[0]
	newVel = np.arange(-Nyquvist[0]*2,Nyquvist[0]*2+dv,dv,dtype=np.float32)
	#print(r)
	datasel = data.sel(range=r)
	#datasel = datasel[['sZDR','HSpec','VSpec','sSNR_H','sSNR_V']]
	sZDR = datasel.sZDR
	HSpec = datasel.HSpec
	VSpec = datasel.VSpec
	sSNR_H = datasel.sSNR_H
	sSNR_V = datasel.sSNR_V
	if r > RangeEnd[1]: # get correct nyquvist vel (I already regridded everything so I need to do it like tha
		maxVel = Nyquvist[2]
		ra = RangeEnd[2]
	elif r <= RangeEnd[0]:
		maxVel = Nyquvist[0]
		ra = RangeEnd[0]
	else:
		maxVel = Nyquvist[1]
		ra = RangeEnd[1]
	
	if (~np.isnan(sZDR.sel(Vel=maxVel,method='nearest')) or ~np.isnan(sZDR.sel(Vel=-maxVel,method='nearest')) or ~np.isnan(sZDR.sel(Vel=-maxVel+dv,method='nearest')) 
		or ~np.isnan(sZDR.sel(Vel=maxVel-dv,method='nearest')) or ~np.isnan(sZDR.sel(Vel=-maxVel+2*dv,method='nearest')) or ~np.isnan(sZDR.sel(Vel=maxVel-2*dv,method='nearest')) 
		or ~np.isnan(sZDR.sel(Vel=-maxVel+3*dv,method='nearest')) or ~np.isnan(sZDR.sel(Vel=maxVel-3*dv,method='nearest'))): 
	# if we have non-noise values at +- Nyquvist range or upto 3 vel bins to the right or left
		datadrop = datasel.where(data.Vel < maxVel,drop=True) #- since we regridded, I need to drop everything that is larger or smaller than +- Ny to actually kit the spectra together (otherwise we would have a lot of nan in between..)
		datadrop = datadrop.where(data.Vel > -maxVel,drop=True)
		q = len(datadrop.Vel.values)-1;
		k = 0;
		
		if np.isnan(datadrop.HSpec.isel(Vel=q)):
			datadrop.HSpec.values[q] = -99999
		for i in range(q):
			if np.isnan(datadrop.HSpec.isel(Vel=[i])) and  ~np.isnan(datadrop.HSpec.isel(Vel=[i+1])):
				datadrop.HSpec.values[i] = -99999
			elif np.isnan(datadrop.HSpec.isel(Vel=[i])) and  ~np.isnan(datadrop.HSpec.isel(Vel=[i+2])):
				datadrop.HSpec.values[i] = -99999
			elif np.isnan(datadrop.HSpec.isel(Vel=[i])) and  ~np.isnan(datadrop.HSpec.isel(Vel=[i+3])):
				datadrop.HSpec.values[i] = -99999
			elif np.isnan(datadrop.HSpec.isel(Vel=[i])) and  ~np.isnan(datadrop.HSpec.isel(Vel=[i+4])):
				datadrop.HSpec.values[i] = -99999
			elif np.isnan(datadrop.HSpec.isel(Vel=[i])) and  ~np.isnan(datadrop.HSpec.isel(Vel=[i+5])):
				datadrop.HSpec.values[i] = -99999
		if datadrop.HSpec.isnull().any(dim='Vel'):
			while ~np.isnan(datadrop.HSpec.isel(Vel=q)): #-- find out how much we need to shift
				k = k + 1;
				q = q - 1;
		# now roll the coordinates:
		datadrop = datadrop.where(datadrop.HSpec!=-99999,np.nan)
		#print(maxVel)
		dataRoll = datadrop.roll(Vel=k+1,roll_coords=False)  # shift everything, the direction doesnt matter
		
		dataRollRegridded = dataRoll.reindex({'Vel':data.Vel.values}, method = 'nearest', tolerance = 0.05) # now I need to go back to original velocity
		HSpec = dataRollRegridded.HSpec
		VSpec = dataRollRegridded.VSpec
		sZDR = dataRollRegridded.sZDR
		sSNR_H = dataRollRegridded.sSNR_H
		sSNR_V = dataRollRegridded.sSNR_V
	
	
	#- now move everything to 0 and name that rot (better for plotting)
	velMatrix = HSpec.values/HSpec.values*vel
	maxVelH = np.nanmax(velMatrix)#,axis=1)
	minVelH = np.nanmin(velMatrix)#,axis=1)
	if np.isnan(maxVelH):
		maxVelH = 0
	if np.isnan(minVelH):
		minVelH = 0
	
	velMatrix = VSpec.values/VSpec.values*vel
	maxVelV = np.nanmax(velMatrix)#,axis=1)
	minVelV = np.nanmin(velMatrix)#,axis=1)
	if np.isnan(maxVelV):
		maxVelV = 0
	if np.isnan(minVelV):
		minVelV = 0
		
	velMatrix = sZDR.values/sZDR.values*vel
	maxVelZDR = np.nanmax(velMatrix)#,axis=1)
	minVelZDR = np.nanmin(velMatrix)#,axis=1)
	if np.isnan(maxVelZDR):
		maxVelZDR = 0
	if np.isnan(minVelZDR):
		minVelZDR = 0
	#- now shift everything to 0, because we might get folding I have to increase Dopplervelocity range to have +-12
	HSpecRe = HSpec.reindex({'Vel':newVel}, method = 'nearest', tolerance = 0.05) 
	vel2rot = int((-maxVelH/dv).round())
	if datasel.maxVelV > 0:
		HSpec_rot = HSpecRe.roll(Vel=vel2rot-1,roll_coords=False)
	else:
		HSpec_rot = HSpecRe.roll(Vel=vel2rot,roll_coords=False)
	
	VSpecRe = VSpec.reindex({'Vel':newVel}, method = 'nearest', tolerance = 0.05) 	
	vel2rot = int((-maxVelV/dv).round())
	if datasel.maxVelV > 0:
		VSpec_rot = VSpecRe.roll(Vel=vel2rot-1,roll_coords=False)
	else:
		VSpec_rot = VSpecRe.roll(Vel=vel2rot,roll_coords=False)
		
	sZDRRe = sZDR.reindex({'Vel':newVel}, method = 'nearest', tolerance = 0.05) 
	vel2rot = int((-maxVelZDR/dv).round())
	if datasel.maxVelZDR > 0:
		sZDR_rot = sZDRRe.roll(Vel=vel2rot-1,roll_coords=False)
	else:
		sZDR_rot = sZDRRe.roll(Vel=vel2rot,roll_coords=False)
	return r,HSpec,VSpec,sZDR,sSNR_H,sSNR_V,maxVelH,minVelH,maxVelV,minVelV,maxVelZDR,minVelZDR,HSpec_rot,VSpec_rot,sZDR_rot

def calcPhiDP(data):
    # change phidp, since it was stored weird
    #PHIDP = data['PhiDP'].values
    
    data['PhiDP'] = np.rad2deg(data['PhiDP']) # convert to deg, add -1 because convention is other way around (now the phase shift gets negative, we want it to get positive with range...) TODO: check with Alexander if that makes sense!!
    #data['PhiDP'] = data['PhiDP'].rolling(range=5, min_periods=1,center=True).mean()
    data['PhiDP'].attrs = {'standard_name':'PhiDP',
                     'long_name': 'Differential phase shift',
                     'units':'deg'}
    return data


def calcKDP(data,timeWindow = 30):
    # time window: timewindow*timeres gives the amount of seconds over which will be averaged 
    # calculate KDP from phidp directly
    delRange = np.diff(data.range)
    if len(data.time)>1: # need to do that because when radar crashed we might have really small files where the time moving mean does not work anymore
      PHIDP = data['PhiDP'].rolling(time=timeWindow,min_periods=1,center=True).mean() # moving window average in time
      PHIDP = PHIDP.rolling(range=5, min_periods=1, center=True).mean() #moving window average in range
      data['KDP'] = PHIDP.diff(dim='range')/(2.*abs(delRange)*1e-3) # in order to get °/km we need to multiply with 1e-3
      data['KDP'].attrs = {'long_name': 'Specific differential phase shift',
                   'units':'deg/km'}
    else:
      data['KDP'] = data['PhiDP'].copy()*np.nan
      data['KDP'].attrs = {'long_name': 'Specific differential phase shift',
                   'units':'deg/km'}  
    return data


import numpy as np
import scipy as sp
import xarray as xr
import dask.array as da
from loguru import logger
import datetime as dt

from gfatpy.lidar.utils.file_manager import channel2info
from gfatpy.utils.utils import moving_average_xarray

""" AD-HOC CORRECTIONS
+ pc peak correction: MULHACEN
"""


def mulhacen_pc_peak_correction(signal):
    """
    Correction of the PC peaks in the PC channels caused by PMT degradation of MULHACEN
    Parameters
    ----------
    signal: array numpy
        lidar signal uncorrected from pc peaks 1D, 2D array (time, range)
    Returns
    -------
    signal_new: array numpy
        lidar signal pc peaks corrected. 2D array (time, range) [numpy.array]
    """

    print("INFO. Start PC Peak Correction")

    # force to 2D
    is_1d = False
    if signal.ndim == 1:
        signal = signal[np.newaxis, :]
        is_1d = True

    # threshold for delta bin:
    threshold = 1000

    new_signal = signal.copy()
    for i in range(signal.shape[0]):
        try:
            profile = np.squeeze(signal[i, :])
            # Obsolete
            # if (profile > 1000).any():
            #    indexes = np.arange(profile.size)
            #    idx_diff = indexes[profile > 200]
            #    for idx_ in idx_diff:
            #        if idx_ < 2:
            #            datacorr = 0.8 * np.median(profile[:10], axis=0)
            #        else:
            #            datacorr = 2 * profile[idx_ - 1] - profile[idx_ - 2]
            #        dif = profile[idx_] - datacorr
            #        profile[idx_] = datacorr
            #        if (idx_ + 1) in idx_diff:
            #            k = 1
            #            while (idx_ + k) in idx_diff:
            #                profile[idx_ + k] = profile[idx_ + k] - dif
            #                k += 1
            #        else:
            #            k = 0
            #        idx_ += k

            diff = np.diff(profile)
            if (diff > threshold).any():
                indexes = np.arange(diff.size)
                idx_diff = indexes[diff > threshold]  # With respect to diff array
                idx_profile = idx_diff + 1  # With respect to profile array
                for idx_ in idx_profile:
                    datacorr = np.mean(profile[idx_ - 3 : idx_ - 1])
                    profile[idx_] = datacorr
                    k = 1
                    while (idx_ + k) < profile.size and (
                        profile[idx_ + k] - profile[idx_]
                    ) > threshold:
                        datacorr = np.mean(profile[idx_ + k - 3 : idx_ + k - 1])
                        profile[idx_ + k] = datacorr
                        k += 1
            profile[
                0:3
            ] = 0  # TODO: Preguntar a JABA el beneficio de poner los primeros bines a 0
            new_signal[i, :] = profile
        except Exception:
            print("pc peak correction not performed for profile %d-th" % i)
    # if 1d:
    if is_1d:
        new_signal = new_signal.ravel()

    print("INFO. End PC Peak Correction")

    return new_signal


""" DARK SIGNAL
"""


def subtract_dark_current(signal, dc):
    """ """
    # set flag default:
    dark_corrected_flag = False
    # same length of range dimension
    do = False
    if dc is not None:
        # DC must be numpy
        dc = dc.astype(np.float64)
        """
        if isinstance(dc, int):
            dc = np.int64(dc)
        elif isinstance(dc, float):
            dc = np.float64(dc)
        """

        if not np.isnan(dc).all():
            if signal.ndim == 1:
                if dc.ndim == 1:
                    if len(signal) == len(dc):
                        do = True
                elif dc.ndim == 0:
                    do = True
                else:
                    print("ERROR. Wrong dimensions")
            elif signal.ndim == 2:
                if dc.ndim == 2:
                    if signal.shape == dc.shape:
                        do = True
                elif dc.ndim == 1:
                    if signal.shape[1] == len(dc):
                        do = True
                elif dc.ndim == 0:
                    do = True
                else:
                    print("ERROR. Wrong dimensions")
            elif signal.ndim == 0:
                if dc.ndim == 0:
                    do = True
            else:
                print("ERROR. Wrong dimensions")

        if do:
            signal = signal - dc
            dark_corrected_flag = True
        else:
            print("WARNING. DC not subtracted")

    return signal, dark_corrected_flag


""" BACKGROUND
"""


def estimate_background(rs, idx_min, idx_max, dc=None):
    """
    Background Signal is estimated by considering values in the very far range,

    Parameters
    ----------
    rs: numpy.array, xarray.Dataarray
        Signal: 1D, 2D array (time, range)
    idx_min: Index of Min Height for Background. int
    idx_max: Index of Max Height for Background. int
    dc: numpy.array, xarray.Dataarray
        DC Signal. 1D array (range)

    Returns
    -------
    bg: Background value. 1D array (time)
    """

    logger.info("Start Estimate Background")
    if rs.ndim == 1:
        rs = rs[np.newaxis, :]
    try:
        # Method: Average RS values over BG height range for each profile
        if dc is None:
            bg = np.nanmean(rs[:, idx_min : idx_max + 1], axis=1)
        else:
            bg = np.nanmean(
                rs[:, idx_min : idx_max + 1] - dc[idx_min : idx_max + 1], axis=1
            )
    except Exception as e:
        bg = None
        logger.error("background cannot be estimated")
        raise RuntimeError(e)
    logger.info("End Estimate Background")

    return bg


def subtract_background(signal, bg):
    """Subtract Background signal to Raw signal

    Parameters
    ----------
    signal : array
        raw signal
    bg : float
        background value

    Returns
    -------
    [type]
        [description]
    """

    do = False

    # BG must be numpy.float64 (it has ndim)
    if isinstance(bg, int):
        bg = np.int64(bg)
    elif isinstance(bg, float):
        bg = np.float64(bg)

    if signal.ndim == 1:
        if bg.ndim == 0:
            do = True
        else:
            print("ERROR. Wrong dimensions")
    elif signal.ndim == 2:
        if bg.ndim == 1:
            if signal.shape[0] == len(bg):
                do = True
        elif bg.ndim == 0:
            do = True
        else:
            print("ERROR. Wrong dimensions")
    elif signal.ndim == 0:
        if bg.ndim == 0:
            do = True
    else:
        print("ERROR. Wrong dimensions")

    if do:
        signal = (signal.T - bg).T
    else:
        print("ERROR. BG not subtracted")

    return signal


def apply_bin_zero_correction(sg, delay):
    """
    TODO: desacoplar dependencia con tipo de variable (numpy/dask array)

    Parameters
    ----------
    sg: array
        signal (range), (time, range)
    delay: float
        position of bin zero (>0, <0)

    Returns
    -------
    sg_c: array
        signal corrected from bin zero

    """

    if not isinstance(delay, int):
        delay = int(delay)

    if isinstance(sg, np.ndarray):
        is_1d = False
        if sg.ndim == 1:
            sg = sg[np.newaxis, :]
            is_1d = True

        sg_c = sg.copy()
        try:
            if delay > 0:
                sg_c[:, :-delay] = sg[:, delay:]
                sg_c[:, -delay:] = np.nan
            elif delay < 0:
                sg_c[:, -delay:] = sg[:, :delay]
                sg_c[:, :-delay] = np.nan
        except Exception as e:
            print("ERROR. In apply_bin_zero_correction. %s" % str(e))
        if is_1d:
            sg_c = sg_c.ravel()
    else:
        if delay > 0:
            sg_c = da.concatenate(
                [sg[:, delay:], da.zeros((sg.shape[0], delay)) * np.nan], axis=1
            )
        elif delay < 0:
            sg_c = da.concatenate(
                [da.zeros((sg.shape[0], abs(delay))) * np.nan, sg[:, :delay]], axis=1
            )
        else:
            sg_c = sg

    return sg_c


""" DEAD TIME
"""


def apply_dead_time_correction(pc, tau, system=0):
    """
    Application of DEAD TIME correction over PC signal

    Parameters
    ----------
    pc: array
        photoncounting signal in MHz
    tau: float
        dead time in ns
    system: int
        paralyzable (1), nonparalyzable (0)

    Returns
    -------
    c: array
        pc signal corrected from dead time. 1D (2D) array (range (time, range))

    """

    try:
        # tau from ns to us
        tau_us = tau * 1e-3
        if system == 0:  # NON-PARALYZABLE
            # Eq 4 [D'Amico et al., 2016]
            c = pc / (1 - pc * tau_us)
        elif system == 1:  # PARALYZABLE
            # To be derived from Eq (2) [D'Amico et al., 2016]. Non-analytic
            c = pc
            print("WARNING: PARALYZABLE NOT IMPLEMENTED. No correction is applied")
        else:
            c = pc
            print("WARNING: wrong system for dead time correction. None is applied")
        # No infinites nor negative values
        c = np.where(np.logical_or(np.isinf(c), c < 0), np.nan, c)

    except Exception as e:
        print("ERROR. In apply_dead_time_correction %s" % str(e))
        c = pc * np.nan

    return c


""" PREPROCESSING SIGNAL
"""


def average_dc_signal(dc):
    """
    Provided a set of DC profiles, a time averaged and vertical smoothed DC profile
    is calculated, which might be broadcasted in a given new time dimension

    Parameters
    ----------
    dc: array (1d, 2d)
        Dark Current (time (dc), range)
    Returns
    -------
    dc_pp: array
        Dark Current Smoothed Profile. 1D (range) or 2D array (time, range)
    """

    logger.info("Start DC Averaging")
    try:
        # force to 2D
        if dc.ndim == 1:  # (range) -> (1, dim_range)
            dc = dc[np.newaxis, :]

        # average dc over time
        dc_avg_t = np.nanmean(dc, axis=0)

        # vertical smoothing of time-averaged profile
        weights = np.array([1, 2, 4, 6, 8, 6, 4, 2, 1], dtype=float)
        weights /= np.sum(weights)
        dc_pp = sp.ndimage.convolve(dc_avg_t, weights, mode="constant")
        dc_pp = np.concatenate([dc_pp[:-4], dc_pp[-8:-4]])
    except Exception as e:
        dc_pp = np.zeros(dc.size)
        logger.error("preprocessing dc. DC set to 0")
        raise RuntimeError(e)

    logger.info("End DC Averaging")
    return dc_pp


def preprocessing_analog_signal(
    signal,
    dc,
    bz,
    idx_min,
    idx_max,
    dc_flag=True,
    zerobin_flag=True,
    bg_flag=True,
    workflow=0,
):
    """

    Parameters
    ----------
    signal: Raw Signal. Measured. 1D, 2D array (time (rs), range)
    dc: Dark Current Signal. Measured. 1D, 2D array (time (dc), range)
    bz: Bin Zero. scalar.
    bg: Background. Pre processed. 1D array (time (rs))
    zerobin_flag: activate/desactivate zero-bin correction. bool
    bg_flag: activate/desactivate background correction. bool
    workflow: type of workflow. Scalar. 0: SCC; 1: BG before ZB

    Returns
    -------
    signal: Preprocessed Signal. 2D array (time (rs), range)

    """

    logger.info("Start Analog Preprocessing")
    try:
        # force to 2D
        is_1d = False
        if signal.ndim == 1:
            signal = signal[np.newaxis, :]
            is_1d = True

        # Type of workflow
        if np.logical_or(workflow < 0, workflow > 1):
            workflow = 0  # SCC

        # Workflow: ZERO_BIN[(RAW - DC)] -  BG
        # 0. Estimate Background
        bg = estimate_background(signal, idx_min, idx_max, dc=dc)

        # _, ax = plt.subplots()
        # ax.plot(signal.mean(axis=0), label='RAW')

        # 1. Subtract DC from AN and BG
        dark_corrected = False
        if dc_flag:
            signal, dark_corrected = subtract_dark_current(signal, dc)
            # dc_signal = np.array(signal)
            # ax.plot(dc_signal.mean(axis=0), label='DC')
        if workflow == 0:  # SCC
            if zerobin_flag:
                # 2. Apply Trigger Delay
                signal = apply_bin_zero_correction(signal, bz)
                # # ax.plot(signal.mean(axis=0), label='BZ')
                # zb_signal = np.array(signal)
            if bg_flag:
                # 3. Subtract Background (which has been subtracted from DC)
                signal = subtract_background(signal, bg)
                # ax.plot(signal.mean(axis=0), label='BG')
                # bg_signal = np.array(signal)
        else:  # Indistinguible de SCC
            if bg_flag:
                # 2. Subtract Background (which has been subtracted from DC)
                signal = subtract_background(signal, bg)
            if zerobin_flag:
                # 3. Apply Trigger Delay
                signal = apply_bin_zero_correction(signal, bz)
        # ax.legend()
        # plt.show()
        # pdb.set_trace()
    except Exception as e:
        logger.critical(str(e))
        logger.critical("Signal not pre-processed.")
        raise RuntimeError("Sinal not preprocessed")

    if is_1d:
        signal = signal.ravel()

    logger.info("End Analog Preprocessing")
    return signal, bg, dark_corrected


def preprocessing_photoncounting_signal(
    signal, tau, bz, idx_min, idx_max, deadtime_flag=True, zerobin_flag=True, workflow=0
):
    """

    Parameters
    ----------
    signal: Raw Signal. 2D array (time (rs), range)
    tau: Dead Time (ns). Scalar
    bz: zero bin for photoncounting (delay_an + bin_shift). Scalar.
    bg: Background. 1D array (time (rs))
    peak_correction: Apply peak correction. Bool
    workflow: type of workflow. Scalar. 0: SCC; 1: BG before ZB

    Returns
    -------
    signal: Preprocessed Signal. 2D array (time (rs), range)
    """

    try:
        # Force to 2D
        is_1d = False
        if signal.ndim == 1:
            signal = signal[np.newaxis, :]
            is_1d = True

        # Type of workflow
        if np.logical_or(workflow < 0, workflow > 1):
            workflow = 0  # SCC

        # Workflow SCC: ZERO_BIN[DT(PK(RAW))] -  BG
        # 0. Estimate Background
        bg = estimate_background(signal, idx_min, idx_max)

        if deadtime_flag:
            # Dead Time Correction
            # 1. Apply Dead Time Correction
            signal = apply_dead_time_correction(signal, tau)

        if workflow == 0:  # SCC
            if zerobin_flag:
                # 2. Apply Trigger Delay
                signal = apply_bin_zero_correction(signal, bz)
            # 3. Subtract Background as estimated from Raw Signal
            signal = subtract_background(signal, bg)
        else:
            # 2. Subtract Background as estimated from Raw Signal
            signal = subtract_background(signal, bg)
            if zerobin_flag:
                # 3. Apply Trigger Delay
                signal = apply_bin_zero_correction(signal, bz)

    except Exception as e:
        print(str(e))
        print("signal not pre-processed.")

    if is_1d:
        signal = signal.ravel()

    return signal, bg


def ff_2D_overlap_from_channels(
    lidar_dataset: xr.Dataset,
    channel_ff: str,
    channel_nf: str,
    norm_range: tuple[float, float] = (2500, 3500),
    min_range: float = 200.,
) -> xr.DataArray:

    #Select window size
    if lidar_dataset.time.size <= 15:
        window_size = 3
    elif lidar_dataset.time.size > 15 and lidar_dataset.time.size <= 30:
        window_size = 5
    elif lidar_dataset.time.size > 30 and lidar_dataset.time.size <= 45:
        window_size = 7
    else:
        window_size = 11


    info_ff = channel2info(channel_ff)
    info_nf = channel2info(channel_nf)

    #Check if channels have the same wavelength
    if info_nf[0] != info_ff[0]:
        raise ValueError(
            f"Channels {channel_ff} and {channel_nf} must have the same wavelength"
        )

    # Select time, range to normalize and range to calculate overlap
    nf = lidar_dataset[f"signal_{channel_nf}"] / lidar_dataset[
        f"signal_{channel_nf}"
    ].sel(range=slice(*norm_range)).mean("range")
    ff = lidar_dataset[f"signal_{channel_ff}"] / lidar_dataset[
        f"signal_{channel_ff}"
    ].sel(range=slice(*norm_range)).mean("range")

    # Make the ratio avoiding division by zero
    overlap_raw = ff.where(nf > 0., other=1.) / nf.where(nf > 0., other=1.)

    #Lower limit    
    overlap = overlap_raw.where(overlap_raw.range > min_range, other=0.)    

    #Find when full overlap is reached
    filtered_overlap = moving_average_xarray(overlap, dim='range', window_size=window_size)
    full_overlap_reached_range = overlap.range[(filtered_overlap > 1.).argmax(dim="range")]

    # Set to 1.0 when full overlap is reached
    overlap = overlap.where(overlap.range <= full_overlap_reached_range, other=1.0)

    #Smooth overlap
    overlap = moving_average_xarray(overlap, dim=('range', 'time'), window_size=(window_size,window_size))

    # Assign attributes
    attrs = ["location", "system"]
    for attr_ in attrs:
        overlap.attrs[attr_] = lidar_dataset.attrs[attr_]

    overlap.attrs["history"] = dt.datetime.now().strftime(
        "Created %a %b %d %H:%M:%S %Y"
    )
    overlap.attrs["wavelength"] = info_nf[0]
    overlap.attrs["channel_ff"] = channel_ff
    overlap.attrs["channel_nf"] = channel_nf

    return overlap



from pathlib import Path
import math
import numpy as np
from datetime import datetime
import pandas as pd

import xarray as xr

from .types import RadarInfoType
from gfatpy.utils.io import read_yaml

""" MODULE For General Lidar Utilities
"""

# RADAR SYSTEM INFO
INFO_FILE = Path(__file__).parent.absolute() / "info.yml"
RADAR_INFO: RadarInfoType = read_yaml(INFO_FILE)

INFO_PLOT_FILE = Path(__file__).parent.absolute() / "plot" / "info.yml"
RADAR_PLOT_INFO = read_yaml(INFO_PLOT_FILE)














# INSERT HERE THE SYSTEM PARAMETERS
general_parameters = {
    "System": "ALHAMBRA",
    "Laser_Pointing_Angle": 3,
    "Molecular_Calc": 0,  # Use ECMWF or US standard atmosphere
    "Latitude_degrees_north": 37.16,
    "Longitude_degrees_east": -3.60,
    "Altitude_meter_asl": 660.0,
    "Call sign": "gr",
}

# LINK YOUR LICEL CHANNELS TO SCC PARAMETERS. USE BT0, BC0 ETC AS NAMES (AS IN LICEL FILES).
channel_parameters = {
    "BT0": {
        "channel_ID": 2203,  # 1064fta
        "channel_string_ID": "1064fta",
        "Background_Low": 100.,
        "Background_High": 800.0,
        "Background_Mode": 0,
        "Laser_Shots": 600,
        "LR_Input": 1,
        "First_Signal_Rangebin": 807,        
        "DAQ_Range": 500.0,
    },
    "BT2": {
        "channel_ID": 2204,  # 532fta
         "channel_string_ID": "532fta",
        "Background_Low": 100.,
        "Background_High": 800.0,
        "Background_Mode": 0,
        "Laser_Shots": 600,
        "LR_Input": 1,
        "First_Signal_Rangebin": 807,
        "DAQ_Range": 100.0,
    },
    "BC2": {
        "channel_ID": 2232,  # 532ftp
        "channel_string_ID": "532ftp",
        "Background_Low": 100.,
        "Background_High": 800.0,
        "Background_Mode": 0,
        "Laser_Shots": 600,
        "LR_Input": 1,
        "First_Signal_Rangebin": 805,
        "DAQ_Range": 100.0,
    },
    "BT4": {
        "channel_ID": 2213,  # 355fpa
        "channel_string_ID": "355fpa",
        "Background_Low": 100.,
        "Background_High": 800.0,
        "Background_Mode": 0,
        "Laser_Shots": 600,
        "LR_Input": 1,
        "First_Signal_Rangebin": 807,
        "DAQ_Range": 100.0,
    },
    "BC4": {
        "channel_ID": 2233,  # 355fpp
        "channel_string_ID": "355fpp",
        "Background_Low": 100.,
        "Background_High": 800.0,
        "Background_Mode": 0,
        "Laser_Shots": 600,
        "LR_Input": 1,
        "First_Signal_Rangebin": 805,
        "DAQ_Range": 100.0,
    },
    "BT5": {
        "channel_ID": 2214,  # 355fsa
        "channel_string_ID": "355fsa",
        "Background_Low": 100.,
        "Background_High": 800.0,
        "Background_Mode": 0,
        "Laser_Shots": 600,
        "LR_Input": 1,
        "First_Signal_Rangebin": 807,
        "DAQ_Range": 100.0,
    },
    "BC5": {
        "channel_ID": 2234,  # 355fsp
        "channel_string_ID": "355fsp",
        "Background_Low": 100.,
        "Background_High": 800.0,
        "Background_Mode": 0,
        "Laser_Shots": 600,
        "LR_Input": 1,
        "First_Signal_Rangebin": 805,
        "DAQ_Range": 100.0,
    },
    
    "BT10": {
        "channel_ID": 2123,  # 1064nta
        "channel_string_ID": "1064nta",
        "Background_Low": 100.,
        "Background_High": 800.0,
        "Background_Mode": 0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "First_Signal_Rangebin": 807,
        "DAQ_Range": 100.0,
    },
    "BT11": {
        "channel_ID": 2124,  # 532npa
        "channel_string_ID": "532npa",
        "Background_Low": 100.,
        "Background_High": 800.0,
        "Background_Mode": 0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "First_Signal_Rangebin": 807,
        "DAQ_Range": 100.0,
    },
    "BC11": {
        "channel_ID": 2268,  # 532npp
        "channel_string_ID": "532npp",
        "Background_Low": 100.,
        "Background_High": 800.0,
        "Background_Mode": 0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "First_Signal_Rangebin": 805,
        "DAQ_Range": 100.0,
    },
    "BT12": {
        "channel_ID": 2126,  # 532nsa
        "channel_string_ID": "532nsa",
        "Background_Low": 100.,
        "Background_High": 800.0,
        "Background_Mode": 0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "First_Signal_Rangebin": 807,
        "DAQ_Range": 100.0,
    },
    "BC12": {
        "channel_ID": 2269,  # 532nsp
        "channel_string_ID": "532nsp",
        "Background_Low": 100.,
        "Background_High": 800.0,
        "Background_Mode": 0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "First_Signal_Rangebin": 805,
        "DAQ_Range": 100.0,
    },
    "BT13": {
        "channel_ID": 2128,  # 355npa
        "channel_string_ID": "355npa",
        "Background_Low": 100.,
        "Background_High": 800.0,
        "Background_Mode": 0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "First_Signal_Rangebin": 807,
        "DAQ_Range": 100.0,
    },
    "BC13": {
        "channel_ID": 2270,  # 355npp
        "channel_string_ID": "355npp",
        "Background_Low": 100.,
        "Background_High": 800.0,
        "Background_Mode": 0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "First_Signal_Rangebin": 805,
        "DAQ_Range": 100.0,
    },
    "BT14": {
        "channel_ID": 2130,  # 355nsa
        "channel_string_ID": "355nsa",
        "Background_Low": 100.,
        "Background_High": 800.0,
        "Background_Mode": 0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "First_Signal_Rangebin": 807,
        "DAQ_Range": 100.0,
    },
    "BC14": {
        "channel_ID": 2271,  # 355nsp
        "channel_string_ID": "355nsp",
        "Background_Low": 100.,
        "Background_High": 800.0,
        "Background_Mode": 0,
        "Laser_Shots": 1200,
        "LR_Input": 1,
        "First_Signal_Rangebin": 805,
        "DAQ_Range": 100.0,
    },  
}